
from typing import List, Callable, Dict

import pyarrow as pa
import copy
import pandas as pd

pd.set_option('display.max_rows', None)  # Show all rows
pd.set_option('display.max_columns', None)  # Show all columns
pd.set_option('display.width', None)  # Auto-detect display width
pd.set_option('display.max_colwidth', None)  # Show full content of each column

from parquetdb.utils import pyarrow_utils


def replace_empty_structs(struct_type, dummy_field=pa.field('dummy_field', pa.int16())):
    dummy_struct_type=pa.struct([dummy_field])
    # If the struct is empty, return the dummy struct
    if len(struct_type)==0:
        return dummy_struct_type

    field_list=[]
    # Iterate over the fields in the struct
    for field in struct_type:
        field_name=field.name
        # Handles the determination of the field type
        if pa.types.is_struct(field.type):
            # Handles empty structs.
            if len(field.flatten())==0:
                field_type=dummy_struct_type
            # Handles nested structs.
            else:
                field_type=replace_empty_structs(field.type)
        else:
            field_type=field.type

        field_list.append((field_name,field_type))

    return pa.struct(field_list)

def replace_empty_structs_in_struct(column_array: pa.Array, dummy_field=pa.field('dummy_field', pa.int16())) -> pa.Array:
    dummy_struct_type=pa.struct([dummy_field])

    # Combine chunks if necessary
    if isinstance(column_array, pa.ChunkedArray):
        column_array = column_array.combine_chunks()

    # Detecting if array is a struct type
    original_type = column_array.type
    if pa.types.is_struct(original_type):
        # Adding dummy field to the struct type
        new_type = replace_empty_structs(original_type)
    else:
        # If the array is not a struct type, return the original array
        return column_array

    original_fields_dict = {field.name: i for i, field in enumerate(original_type)}

    # Build a new struct with all the required fields
    new_arrays=[]
    for field in new_type:
        if field.name in original_fields_dict:
            print("Adding values to a existing field")
            # Recursively generate the new array for the field
            field_array = column_array.field(original_fields_dict[field.name])
            new_field_array = replace_empty_structs_in_struct(field_array)
            new_arrays.append(new_field_array)
        else:
            print("Adding null values to a previously non-existing field")
            null_array = pa.nulls(len(column_array), field.type)
            new_arrays.append(null_array)
    return pa.StructArray.from_arrays(new_arrays, fields=new_type)

def replace_empty_structs_in_column(column_array, field, dummy_type=pa.field('dummy_field', pa.int16())):    
    column_type = column_array.type
    print("Field name: ", field.name)
    print("Column type: ", column_type)
    print("Dummy_type type: ", dummy_type)
    if pa.types.is_struct(column_type):
        print('This column is a struct')
        # Replacing empty structs with dummy structs
        new_array=replace_empty_structs_in_struct(column_array)
        new_field=pa.field(field.name, new_array.type)
        return new_array, new_field
    else:
        print('This column is not a struct')
        # If the column is not a struct type, return the original column
        return column_array, field

def replace_empty_structs_in_table(table, dummy_type=pa.field('dummy_field', pa.int16())):
    for col_idx in range(table.num_columns):
        field=table.field(col_idx)
        column_array = table.column(col_idx)

        column_array, field = replace_empty_structs_in_column(column_array, field, dummy_type=dummy_type)
        table = table.set_column(col_idx, field, column_array)
    return table

def add_new_null_fields_in_struct(column_array, new_struct_type):
    # Combine chunks if necessary
    if isinstance(column_array, pa.ChunkedArray):
        column_array = column_array.combine_chunks()

    # Detecting if array is a struct type
    original_type = column_array.type
    if not pa.types.is_struct(original_type):
        return column_array

    original_fields_dict = {field.name: i for i, field in enumerate(original_type)}

    new_arrays=[]
    for field in new_struct_type:
        if field.name in original_fields_dict:
            print("Adding values to a existing field")
            # Recursively generate the new array for the field
            field_array = column_array.field(original_fields_dict[field.name])
            new_field_array = add_new_null_fields_in_struct(field_array, field_array.type)
            new_arrays.append(new_field_array)
        else:
            print("Adding null values to a previously non-existing field")
            null_array = pa.nulls(len(column_array), field.type)
            new_arrays.append(null_array)
    return pa.StructArray.from_arrays(new_arrays, fields=new_struct_type)

def add_new_null_fields_in_column(column_array, field, new_type):    
    column_type = column_array.type
    print("Field name: ", field.name)
    print("Column type: ", column_type)
    print("New type: ", new_type)
    
    if pa.types.is_struct(column_type):
        print('This column is a struct')
        # Replacing empty structs with dummy structs
        new_type_names=[field.name for field in new_type]
        if field.name in new_type_names:
            new_struct_type=new_type.field(field.name).type
        else:
            new_struct_type=new_type
        new_struct_type = pyarrow_utils.merge_structs(new_struct_type,column_type)
        print("New struct type: ", new_struct_type)
        new_array=add_new_null_fields_in_struct(column_array, new_struct_type)
        new_field=pa.field(field.name, new_array.type)
        return new_array, new_field
    else:
        print('This column is not a struct')
        return column_array, field

def add_new_null_fields_in_table(table, new_schema):
    new_columns_fields=[]
    new_columns=[]
    for field in new_schema:
        if field.name not in table.schema.names:
            new_column=pa.nulls(table.num_rows, type=field.type)
            new_field=pa.field(field.name, field.type)  
        else:
            original_column = table.column(field.name)
            new_column, new_field = add_new_null_fields_in_column(original_column, field, field.type)

        new_columns.append(new_column)
        new_columns_fields.append(new_field)
    table = pa.Table.from_arrays(new_columns, schema=pa.schema(new_columns_fields))
    return table

def order_fields_in_struct(column_array, new_struct_type):
    # Combine chunks if necessary
    if isinstance(column_array, pa.ChunkedArray):
        column_array = column_array.combine_chunks()

    # Detecting if array is a struct type
    original_type = column_array.type
    if not pa.types.is_struct(original_type):
        return column_array

    new_arrays = []
    for field in new_struct_type:
        
        field_array = column_array.field(field.name)
        child_type = new_struct_type.field(field.name).type

        if pa.types.is_struct(field.type):
            new_field_array = order_fields_in_struct(field_array, child_type)
        else:
            new_field_array = field_array

        new_arrays.append(new_field_array)


    return pa.StructArray.from_arrays(new_arrays, fields=new_struct_type)

def order_fields_in_column(column_array, field):
    # Combine chunks if necessary
    if pa.types.is_struct(field.type):
        new_struct_type = field.type
        column_array = order_fields_in_struct(column_array, new_struct_type)
    else:
        column_array = column_array
    return column_array

def order_fields_in_table(table, new_schema):
    new_columns = []
    for field in new_schema:
        original_column = table.column(field.name)
        column_array=order_fields_in_column(original_column, field)
        new_columns.append(column_array)

    return pa.Table.from_arrays(new_columns, schema=new_schema)

def merge_tables(current_table: pa.Table, incoming_table: pa.Table) -> pa.Table:
    """
    Combines the current table and incoming table by aligning them to the merged schema and filling missing fields with nulls.

    Args:
        current_table (pa.Table): The current table.
        incoming_table (pa.Table): The incoming table.
        merged_schema (pa.Schema): The merged schema.

    Returns:
        pa.Table: The combined table.
    """
    current_table=replace_empty_structs_in_table(current_table)
    incoming_table=replace_empty_structs_in_table(incoming_table)

    current_schema = current_table.schema
    incoming_schema = incoming_table.schema
    merged_schema = pyarrow_utils.merge_schemas(current_schema, incoming_schema)

    current_table=add_new_null_fields_in_table(current_table, merged_schema)
    incoming_table=add_new_null_fields_in_table(incoming_table, merged_schema)

    current_table=order_fields_in_table(current_table, merged_schema)
    incoming_table=order_fields_in_table(incoming_table, merged_schema)

    # # Combine the tables
    combined_table = pa.concat_tables([current_table, incoming_table])
    
    return combined_table

if __name__ == "__main__":

    
    data_1 = [
            {'a': 1, 'b': {'x': {'x': {}, 'y': 20}, 'y': 20}, 'c': {}},
            {'a': 2, 'b': {'x': {'x': {}, 'y': 20}, 'y': 40}, 'c': {}},
        ]

    data_2 = [
            { 'b': {'z': 20}, 'd':1}
        ]


    # data = [
    #         { 'c':  {'x': {'x': {}, 'y': 20} }},
    #         {'c':  {'x': {'x': {}, 'y': 20} }},
    #     ]
    # print(data)
    table_1=pa.Table.from_pylist(data_1)
    table_2=pa.Table.from_pylist(data_2)

    # new_schema=pyarrow_utils.merge_schemas(table_1.schema, table_2.schema)
    # print(new_schema)
    # table_1=replace_empty_structs_in_table(table_1)

    # table_1=add_new_null_fields_in_table(table_1, new_schema)
    # df_1 = table_1.to_pandas()
    # print(df_1.columns)
    # print(df_1.head())

    merged_table=merge_tables(table_1, table_2)

    df_merged = merged_table.to_pandas()
    print(df_merged.columns)
    print(df_merged.head())



import logging
import unittest
import shutil
import os
import tempfile

import numpy as np
from parquetdb import ParquetDatasetDB
import pyarrow as pa
import pyarrow.compute as pc
import pandas as pd

# TODO: Create tests for nested structure updates
# TODO: Create tests for 

class TestParquetDatasetDB(unittest.TestCase):
    def setUp(self):
        # Create a temporary directory for the database
        self.temp_dir = tempfile.mkdtemp()
        self.dataset_name='test_dataset'
        self.db = ParquetDatasetDB(dataset_name=self.dataset_name, dir=self.temp_dir, n_cores=1)

        # Create some test data
        self.test_data = [
            {'name': 'Alice', 'age': 30},
            {'name': 'Bob', 'age': 25},
            {'name': 'Charlie', 'age': 35}
        ]
        self.test_df = pd.DataFrame(self.test_data)

    def tearDown(self):
        # Remove the temporary directory after the test
        if os.path.exists(self.temp_dir):
            shutil.rmtree(self.temp_dir)

    def test_create_and_read(self):
        # Test creating data and reading it back
        data = [
            {'name': 'Alice', 'age': 30},
            {'name': 'Bob', 'age': 25}
        ]
        self.db.create(data)

        # Read the data back
        result = self.db.read()
        df = result.to_pandas()

        # Assertions
        self.assertEqual(len(df), 2)
        self.assertIn('name', df.columns)
        self.assertIn('age', df.columns)
        self.assertEqual(df.iloc[0]['name'], 'Alice')
        self.assertEqual(df.iloc[1]['name'], 'Bob')

    def test_update(self):
        # Test updating existing records

        data = [
            {'name': 'Charlie', 'age': 28},
            {'name': 'Diana', 'age': 32}
        ]
        self.db.create(data)

        # Update the age of 'Charlie'
        update_data = [
            {'id': 0, 'age': 29}
        ]
        self.db.update(update_data)

        # Read the data back
        result = self.db.read()
        df = result.to_pandas()

        # Assertions
        self.assertEqual(df.iloc[0]['age'], 29)
        self.assertEqual(df.iloc[1]['age'], 32)

    def test_delete(self):
        # Test deleting records
        data = [
            {'name': 'Eve', 'age': 35},
            {'name': 'Frank', 'age': 40}
        ]
        self.db.create(data, )

        # Delete 'Eve'
        self.db.delete(ids=[0], )

        # Read the data back
        result = self.db.read()
        df = result.to_pandas()

        # Assertions
        self.assertEqual(len(df), 1)
        self.assertEqual(df.iloc[0]['name'], 'Frank')

    def test_filters(self):
        # Test reading data with filters
        data = [
            {'name': 'Grace', 'age': 22},
            {'name': 'Heidi', 'age': 27},
            {'name': 'Ivan', 'age': 35}
        ]
        self.db.create(data)

        # Apply filter to get people older than 25
        age_filter = pc.field('age') > 25
        result = self.db.read(filters=[age_filter])
        df = result.to_pandas()

        # Assertions
        self.assertEqual(len(df), 2)
        self.assertListEqual(df['name'].tolist(), ['Heidi', 'Ivan'])

    def test_add_new_field(self):
        # Test adding data with a new field and ensure schema evolves
        # This also test that if new incoming data doesn't have a 
        # field that is already in the schema, a null value of the correct type is added
        data = [
            {'name': 'Judy', 'age': 29}
        ]
        self.db.create(data)

        # Add new data with an additional field
        new_data = [
            {'name': 'Karl', 'occupation': 'Engineer'}
        ]
        self.db.create(new_data)

        # Read back the data
        result = self.db.read()
        df = result.to_pandas()

        # Assertions
        self.assertIn('occupation', df.columns)
        self.assertEqual(df.iloc[1]['occupation'], 'Engineer')
        self.assertTrue(pd.isnull(df.iloc[0]['occupation']))
        self.assertTrue(np.isnan(df.iloc[1]['age']))


    def test_nested_empty_struct_field(self):
        # Test adding data with a new field and ensure schema evolves
        # This also test that if new incoming data doesn't have a 
        # field that is already in the schema, a null value of the correct type is added
        data = [
            {'name': {}, 'age': {"field1": {"inner_field1": 1, "inner_field2": 2}}},
        ]
        self.db.create(data)

        # Add new data with an additional field
        new_data = [
            {'name': {'age':1}, 'occupation': 'Engineer'}
        ]
        self.db.create(new_data)

        # Read back the data
        result = self.db.read()
        df = result.to_pandas()

        # Assertions
        # self.assertIn('occupation', df.columns)
        self.assertIn('dummy_field', df.iloc[0]['name'])
        self.assertEqual(df.iloc[0]['name']['dummy_field'], None)
        self.assertEqual(df.iloc[1]['name']['dummy_field'], None)
        self.assertEqual(df.iloc[0]['age']['field1']['inner_field1'], 1)
        self.assertEqual(df.iloc[1]['name']['age'], 1)

    def test_get_schema(self):
        # Test retrieving the schema
        data = [
            {'name': 'Liam', 'age': 45}
        ]
        self.db.create(data, )
        schema = self.db.get_schema()

        # Assertions
        self.assertIn('name', schema.names)
        self.assertIn('age', schema.names)
        self.assertIn('id', schema.names)

    def test_read_specific_columns(self):
        # Test reading specific columns
        data = [
            {'name': 'Mia', 'age': 30, 'city': 'New York'},
            {'name': 'Noah', 'age': 35, 'city': 'San Francisco'}
        ]
        self.db.create(data)

        # Read only the 'name' column
        result = self.db.read(columns=['name'])
        df = result.to_pandas()

        # Assertions
        self.assertEqual(len(df.columns), 1)
        self.assertIn('name', df.columns)
        self.assertNotIn('age', df.columns)
        self.assertNotIn('city', df.columns)

    def test_batch_reading(self):
        # Test reading data in batches
        data = [{'name': f'Person {i}', 'age': i} for i in range(100)]
        self.db.create(data)

        # Read data in batches of 20
        batches = self.db.read(batch_size=20, output_format='batch_generator')

        # Assertions
        batch_count = 0
        total_rows = 0
        for batch in batches:
            batch_count += 1
            total_rows += batch.num_rows
            self.assertLessEqual(batch.num_rows, 20)
        self.assertEqual(batch_count, 5)
        self.assertEqual(total_rows, 100)

    def test_update_schema(self):
        # Test updating the schema of the table
        data = [
            {'name': 'Olivia', 'age': 29}
        ]
        self.db.create(data)
        
        # Update the 'age' field to be a float instead of int
        new_field = pa.field('age', pa.float64())
        field_dict = {'age': new_field}
        self.db.update_schema(field_dict=field_dict)

        # Read back the data
        result = self.db.read()
        df = result.to_pandas()

        # Assertions
        self.assertEqual(df['age'].dtype, 'float64')

    def test_update_with_new_field_included(self):
        # Test updating the schema of the table
        data = [
            {'name': 'Mia', 'age': 30, 'city': 'New York'},
            {'name': 'Noah', 'age': 35, 'city': 'San Francisco'}
        ]
        self.db.create(data)

        # Update the 'Mia' record to include a new field and change age to 60
        data = {'id':0, 'age': 60, 'state':'NY'}
        self.db.update(data,)

        # Read back the data
        result = self.db.read()
        df = result.to_pandas()

        # Assertions
        # print(df.head())
        self.assertEqual(df.iloc[0]['state'], 'NY')
        self.assertEqual(df.iloc[1]['state'], None)
        self.assertEqual(df.iloc[0]['age'], 60)
        self.assertEqual(df.iloc[1]['age'], 35)

    def test_invalid_dataset_name(self):
        # Test using a reserved table name
        with self.assertRaises(ValueError):
            self.db.create(data=[], dataset_name='tmp')

    def test_delete_nonexistent_id(self):
        # Test deleting an ID that doesn't exist
        data = [
            {'name': 'Peter', 'age': 50}
        ]
        self.db.create(data)

        # Attempt to delete a non-existent ID
        self.db.delete(ids=[999])

        # Read back the data
        result = self.db.read()
        df = result.to_pandas()

        # Assertions
        self.assertEqual(len(df), 1)
        self.assertEqual(df.iloc[0]['name'], 'Peter')

    def test_update_nonexistent_id(self):
        # Test updating an ID that doesn't exist
        data = [
            {'name': 'Quinn', 'age': 40}
        ]
        self.db.create(data)
        
        # Attempt to update a non-existent ID
        update_data = [
            {'id': 999, 'age': 41}
        ]
        with self.assertRaises(ValueError):
            self.db.update(update_data)

    def test_get_metadata(self):
        self.db.create(data=self.test_data,
                       metadata={'key1':'value1', 'key2':'value2'})
        # Should return metadata dictionary (can be empty)
        metadata = self.db.get_metadata()
        self.assertIsInstance(metadata, dict)


    def test_drop_dataset(self):
        self.db.create(data=self.test_data)
        # Drop the table and check if it no longer exists
        self.db.drop_dataset()


    def test_rename_dataset(self):
        self.db.create(data=self.test_data)
        # Rename the table and check if the new name exists
        self.db.rename_dataset('renamed_table')

        # Attempt to rename to a reserved name
        with self.assertRaises(ValueError):
            self.db.rename_dataset('tmp')

    def test_export_dataset(self):
        self.db.create(data=self.test_data)
        # Export the table to CSV
        export_path = os.path.join(self.temp_dir, 'exported_table.csv')
        self.db.export_dataset(export_path, format='csv')
        self.assertTrue(os.path.exists(export_path))

        # Verify the exported data
        exported_df = pd.read_csv(export_path)
        original_df = self.db.read().to_pandas()
        pd.testing.assert_frame_equal(original_df, exported_df)

        # Export to an unsupported format
        with self.assertRaises(ValueError):
            self.db.export_dataset(export_path, format='xlsx')

    # def test_merge_datasets(self):
    #     self.db.create(data=self.test_data)
    #     # Create another table
    #     additional_data = [
    #         {'id': 4, 'name': 'Dave', 'age': 40},
    #         {'id': 5, 'name': 'Eve', 'age': 45}
    #     ]
    #     self.db.create(data=additional_data, dataset_name='additional_table')

    #     # Attempt to merge tables (method not implemented)
    #     with self.assertRaises(NotImplementedError):
    #         self.db.merge_datasets(['test_table', 'additional_table'], 'merged_table')

    # def test_deep_update(self):
    #     original_value = {'a': 1, 'b': {'c': 2, 'd': 3}}
    #     update_value = {'a': 10, 'b': {'c': 20, 'e': 30}}
    #     expected_value = {'a': 10, 'b': {'c': 20, 'd': 3, 'e': 30}}
    #     result = deep_update(original_value, update_value)
    #     self.assertEqual(result, expected_value)

if __name__ == '__main__':
    unittest.main()



import pyarrow as pa
import logging

logger = logging.getLogger(__name__)

# https://arrow.apache.org/docs/python/api/datatypes.html
t_string=pa.string()
t_int32=pa.int32()
t_int64=pa.int64()
t_float32=pa.float32()
t_float64=pa.float64()
t_bool=pa.bool_()

# Create variable-length or fixed size binary type.
t_binary = pa.binary()

#one of ‘s’ [second], ‘ms’ [millisecond], ‘us’ [microsecond], or ‘ns’ [nanosecond]
t_timestamp=pa.timestamp('ms')


def find_difference_between_pyarrow_schemas(schema1, schema2):
    """
    Finds the difference between two PyArrow schemas.

    Args:
        schema1 (pyarrow.Schema): The first schema.
        schema2 (pyarrow.Schema): The second schema.

    Returns:
        set: A set of field names that are in schema1 but not in schema2.
    """
    # Create a set of field names from the first schema
    field_names1 = set(schema1)
    # Create a set of field names from the second schema
    field_names2 = set(schema2)
    # Find the difference between the two sets
    difference = field_names1.difference(field_names2)
    return difference

def merge_structs(current_type: pa.StructType, incoming_type: pa.StructType) -> pa.StructType:
    # Create a dictionary of the current fields for easy comparison
    current_fields_dict = {field.name: field for field in current_type}
    merged_fields = []

    # Iterate over the incoming fields and check if they exist in the current_type
    for incoming_field in incoming_type:
        if incoming_field.name in current_fields_dict:
            current_field = current_fields_dict[incoming_field.name]

            # If both are structs, recursively merge the fields
            if pa.types.is_struct(incoming_field.type) and pa.types.is_struct(current_field.type):
                merged_field = pa.field(
                    incoming_field.name,
                    merge_structs(current_field.type, incoming_field.type)
                )
            else:
                # If the field exists but has a different type, use the incoming field type
                merged_field = incoming_field
        else:
            # If the field does not exist in the current, add the incoming field
            merged_field = incoming_field

        merged_fields.append(merged_field)

    # Add any remaining current fields that were not in the incoming schema
    for current_field in current_type:
        if current_field.name not in {field.name for field in incoming_type}:
            merged_fields.append(current_field)

    return pa.struct(merged_fields)

def schema_to_struct(schema):
    return pa.struct(schema)

def merge_schemas(current_schema: pa.Schema, incoming_schema: pa.Schema) -> pa.Schema:
    
    merged_fields = []
    incoming_field_names = {field.name for field in incoming_schema}
    # Iterate through fields in the current schema
    for current_field in current_schema:
        if current_field.name in incoming_field_names:
            incoming_field = incoming_schema.field(current_field.name)
    
            if pa.types.is_struct(current_field.type) and pa.types.is_struct(incoming_field.type):
                # If both fields are structs, merge them recursively
                merged_field = pa.field(
                    current_field.name,
                    merge_structs(current_field.type, incoming_field.type)
                )
            else:
                # Use the incoming field if they are not both structs or if they are identical
                merged_field = incoming_field
        else:
            # If the field does not exist in the incoming schema, keep the current field
            merged_field = current_field
        
        merged_fields.append(merged_field)

    # Add any fields from the incoming schema that are not in the current
    for incoming_field in incoming_schema:
        if incoming_field.name not in {field.name for field in current_schema}:
            merged_fields.append(incoming_field)

    # Return the merged schema
    return pa.schema(merged_fields)

def add_null_columns_for_missing_fields(table: pa.Table, new_schema: pa.Schema) -> pa.Table:
    """
    Adds null columns for any fields that are missing in the new schema.

    Args:
        table (pa.Table): The table to add null columns to.
        new_schema (pa.Schema): The schema of the new table.

    Returns:
        pa.Table: The table with null columns added for missing fields.
    """
    table_schema = table.schema
    field_names_missing = list(set(new_schema.names) - set(table_schema.names))
    
    # Add missing fields with null values
    for new_field_name in field_names_missing:
        field_type = new_schema.field(new_field_name).type
        null_array = pa.nulls(table.num_rows, type=field_type)
        table = table.append_column(new_field_name, null_array)
    
    return table

def align_struct_fields(original_array: pa.Array, new_type: pa.DataType, dummy_type=pa.struct([('dummy_field', pa.int16())])) -> pa.Array:
    """
    Aligns the fields of an array to match the new type, filling missing fields with null values.
    Handles nested struct types recursively and replaces empty structs with a dummy struct.

    Args:
        original_array (pa.Array): The original array (can be a struct or any other type).
        new_type (pa.DataType): The target type to align to.

    Returns:
        pa.Array: The aligned array.
    """
    
    # Combine chunks if necessary
    if isinstance(original_array, pa.ChunkedArray):
        original_array = pa.concat_arrays(original_array.chunks)
    
    # If the new type is not a struct, just cast and return
    if not isinstance(new_type, pa.StructType):
        return original_array.cast(new_type)

    dummy_field_name = [field.name for field in dummy_type][-1]

    original_type = original_array.type
    original_fields_dict = {field.name: i for i, field in enumerate(original_type)}
    
    # Build a new struct with all the required fields
    new_arrays=[]
    for field in new_type:
        if field.name in original_fields_dict:
            # If the field exists, align it recursively
            original_field = original_array.field(original_fields_dict[field.name])
            aligned_field = align_struct_fields(original_field, field.type)
            
            # Check if the aligned field is an empty struct
            if isinstance(field.type, pa.StructType) and len(field.type) == 0:
                # Replace empty struct with dummy struct
                dummy_array = pa.array([{dummy_field_name: None}] * len(original_array), type=dummy_type)
                new_arrays.append(dummy_array)
            else:
                new_arrays.append(aligned_field)
        else:
            # If the field doesn't exist, fill with nulls
            if isinstance(field.type, pa.StructType) and len(field.type) == 0:
                # For empty struct fields, use the dummy struct
                dummy_array = pa.array([{dummy_field_name: None}] * len(original_array), type=dummy_type)
                new_arrays.append(dummy_array)
            else:
                null_array = pa.nulls(len(original_array), field.type)
                new_arrays.append(null_array)
    
    return pa.StructArray.from_arrays(new_arrays, fields=new_type)

def align_table(current_table: pa.Table, new_schema: pa.Schema) -> pa.Table:
    """
    Aligns the given table to the new schema, filling in missing fields or struct fields with null values.

    Args:
        table (pa.Table): The table to align.
        new_schema (pa.Schema): The target schema to align the table to.

    Returns:
        pa.Table: The aligned table.
    """
    current_table=replace_empty_structs_in_table(current_table)

    current_table=add_new_null_fields_in_table(current_table, new_schema)
    current_table=order_fields_in_table(current_table, new_schema)
    
    return current_table

def combine_tables(current_table: pa.Table, incoming_table: pa.Table, merged_schema: pa.Schema=None) -> pa.Table:
    """
    Combines the current table and incoming table by aligning them to the merged schema and filling missing fields with nulls.

    Args:
        current_table (pa.Table): The current table.
        incoming_table (pa.Table): The incoming table.
        merged_schema (pa.Schema): The merged schema.

    Returns:
        pa.Table: The combined table.
    """
    if merged_schema is None:
        current_schema = current_table.schema
        incoming_schema = incoming_table.schema
        merged_schema = merge_schemas(current_schema, incoming_schema)

    current_table=align_table(current_table, merged_schema)
    incoming_table=align_table(incoming_table, merged_schema)

    # Reorder the columns to match the merged schema
    current_table = current_table.select(merged_schema.names)
    incoming_table = incoming_table.select(merged_schema.names)

    # Combine the tables
    combined_table = pa.concat_tables([current_table, incoming_table])
    
    return combined_table

def replace_none_with_nulls(data, schema_field):
    """
    Replaces None values in the data list with null values according to the schema.

    Args:
        data (list): A list of elements to check.
        schema_field (pa.Field or pa.DataType): The field or type that defines the schema.

    Returns:
        list: The updated list with None values replaced with nulls.
    """
    # Check if the schema_field is a pa.Field or pa.DataType
    schema_type = schema_field.type if isinstance(schema_field, pa.Field) else schema_field

    updated_data = []

    for element in data:
        if element is None:
            # Replace None with the equivalent null value
            null_value = pa.scalar(None, type=schema_type)
            updated_data.append(null_value.as_py())
        elif pa.types.is_struct(schema_type):
            # If the field is a struct, recursively check its fields
            updated_data.append(
                {key: replace_none_with_nulls([val], schema_type.field(key))[0] 
                 if val is None else val for key, val in element.items()}
            )
        elif pa.types.is_list(schema_type):
            # If the field is a list, recursively check the elements of the list
            updated_data.append(replace_none_with_nulls(element, schema_type.value_type))
        else:
            # Otherwise, keep the element as is
            updated_data.append(element)

    return updated_data



def replace_empty_structs(struct_type, dummy_field=pa.field('dummy_field', pa.int16())):
    dummy_struct_type=pa.struct([dummy_field])
    # If the struct is empty, return the dummy struct
    if len(struct_type)==0:
        return dummy_struct_type

    field_list=[]
    # Iterate over the fields in the struct
    for field in struct_type:
        field_name=field.name
        # Handles the determination of the field type
        if pa.types.is_struct(field.type):
            # Handles empty structs.
            if len(field.flatten())==0:
                field_type=dummy_struct_type
            # Handles nested structs.
            else:
                field_type=replace_empty_structs(field.type)
        else:
            field_type=field.type

        field_list.append((field_name,field_type))

    return pa.struct(field_list)

def replace_empty_structs_in_struct(column_array: pa.Array, dummy_field=pa.field('dummy_field', pa.int16())) -> pa.Array:
    dummy_struct_type=pa.struct([dummy_field])

    # Combine chunks if necessary
    if isinstance(column_array, pa.ChunkedArray):
        column_array = column_array.combine_chunks()

    # Detecting if array is a struct type
    original_type = column_array.type
    if pa.types.is_struct(original_type):
        # Adding dummy field to the struct type
        new_type = replace_empty_structs(original_type)
    else:
        # If the array is not a struct type, return the original array
        return column_array

    original_fields_dict = {field.name: i for i, field in enumerate(original_type)}

    # Build a new struct with all the required fields
    new_arrays=[]
    for field in new_type:
        if field.name in original_fields_dict:
            logger.debug("Adding values to a existing field")
            # Recursively generate the new array for the field
            field_array = column_array.field(original_fields_dict[field.name])
            new_field_array = replace_empty_structs_in_struct(field_array)
            new_arrays.append(new_field_array)
        else:
            logger.debug("Adding null values to a previously non-existing field")
            null_array = pa.nulls(len(column_array), field.type)
            new_arrays.append(null_array)
    return pa.StructArray.from_arrays(new_arrays, fields=new_type)

def replace_empty_structs_in_column(column_array, field, dummy_type=pa.field('dummy_field', pa.int16())):    
    column_type = column_array.type
    logger.debug(f"Field name: {field.name}")
    logger.debug(f"Column type: {column_type}")
    logger.debug(f"Dummy_type type: {dummy_type}")
    if pa.types.is_struct(column_type):
        logger.debug('This column is a struct')
        # Replacing empty structs with dummy structs
        new_array=replace_empty_structs_in_struct(column_array)
        new_field=pa.field(field.name, new_array.type)
        return new_array, new_field
    else:
        logger.debug('This column is not a struct')
        # If the column is not a struct type, return the original column
        return column_array, field

def replace_empty_structs_in_table(table, dummy_type=pa.field('dummy_field', pa.int16())):
    for col_idx in range(table.num_columns):
        field=table.field(col_idx)
        column_array = table.column(col_idx)

        column_array, field = replace_empty_structs_in_column(column_array, field, dummy_type=dummy_type)
        table = table.set_column(col_idx, field, column_array)
    return table

def add_new_null_fields_in_struct(column_array, new_struct_type):
    # Combine chunks if necessary
    if isinstance(column_array, pa.ChunkedArray):
        column_array = column_array.combine_chunks()

    # Detecting if array is a struct type
    original_type = column_array.type
    if not pa.types.is_struct(original_type):
        return column_array

    original_fields_dict = {field.name: i for i, field in enumerate(original_type)}

    new_arrays=[]
    for field in new_struct_type:
        if field.name in original_fields_dict:
            logger.debug("Adding values to a existing field")
            # Recursively generate the new array for the field
            field_array = column_array.field(original_fields_dict[field.name])
            new_field_array = add_new_null_fields_in_struct(field_array, field_array.type)
            new_arrays.append(new_field_array)
        else:
            logger.debug("Adding null values to a previously non-existing field")
            null_array = pa.nulls(len(column_array), field.type)
            new_arrays.append(null_array)
    return pa.StructArray.from_arrays(new_arrays, fields=new_struct_type)

def add_new_null_fields_in_column(column_array, field, new_type):    
    column_type = column_array.type
    logger.debug(f"Field name:  {field.name}")
    logger.debug(f"Column type: {column_type}")
    logger.debug(f"New type: {new_type}")
    
    if pa.types.is_struct(column_type):
        logger.debug('This column is a struct')
        # Replacing empty structs with dummy structs
        new_type_names=[field.name for field in new_type]
        if field.name in new_type_names:
            new_struct_type=new_type.field(field.name).type
        else:
            new_struct_type=new_type
        new_struct_type = merge_structs(new_struct_type,column_type)
        logger.debug(f"New struct type: {new_struct_type}")
        new_array=add_new_null_fields_in_struct(column_array, new_struct_type)
        new_field=pa.field(field.name, new_array.type)
        return new_array, new_field
    else:
        logger.debug('This column is not a struct')
        return column_array, field

def add_new_null_fields_in_table(table, new_schema):
    new_columns_fields=[]
    new_columns=[]
    for field in new_schema:
        if field.name not in table.schema.names:
            new_column=pa.nulls(table.num_rows, type=field.type)
            new_field=pa.field(field.name, field.type)  
        else:
            original_column = table.column(field.name)
            new_column, new_field = add_new_null_fields_in_column(original_column, field, field.type)

        new_columns.append(new_column)
        new_columns_fields.append(new_field)
    table = pa.Table.from_arrays(new_columns, schema=pa.schema(new_columns_fields))
    return table

def order_fields_in_struct(column_array, new_struct_type):
    # Combine chunks if necessary
    if isinstance(column_array, pa.ChunkedArray):
        column_array = column_array.combine_chunks()

    # Detecting if array is a struct type
    original_type = column_array.type
    if not pa.types.is_struct(original_type):
        return column_array

    new_arrays = []
    for field in new_struct_type:
        
        field_array = column_array.field(field.name)
        child_type = new_struct_type.field(field.name).type

        if pa.types.is_struct(field.type):
            new_field_array = order_fields_in_struct(field_array, child_type)
        else:
            new_field_array = field_array

        new_arrays.append(new_field_array)


    return pa.StructArray.from_arrays(new_arrays, fields=new_struct_type)

def order_fields_in_column(column_array, field):
    # Combine chunks if necessary
    if pa.types.is_struct(field.type):
        new_struct_type = field.type
        column_array = order_fields_in_struct(column_array, new_struct_type)
    else:
        column_array = column_array
    return column_array

def order_fields_in_table(table, new_schema):
    new_columns = []
    for field in new_schema:
        original_column = table.column(field.name)
        column_array=order_fields_in_column(original_column, field)
        new_columns.append(column_array)

    return pa.Table.from_arrays(new_columns, schema=new_schema)

def merge_tables(current_table: pa.Table, incoming_table: pa.Table, schema=None) -> pa.Table:
    """
    Combines the current table and incoming table by aligning them to the merged schema and filling missing fields with nulls.

    Args:
        current_table (pa.Table): The current table.
        incoming_table (pa.Table): The incoming table.
        merged_schema (pa.Schema): The merged schema.

    Returns:
        pa.Table: The combined table.
    """
    current_table=replace_empty_structs_in_table(current_table)
    incoming_table=replace_empty_structs_in_table(incoming_table)

    if schema is None:
        current_schema = current_table.schema
        incoming_schema = incoming_table.schema
        merged_schema = merge_schemas(current_schema, incoming_schema)

    current_table=add_new_null_fields_in_table(current_table, merged_schema)
    incoming_table=add_new_null_fields_in_table(incoming_table, merged_schema)

    current_table=order_fields_in_table(current_table, merged_schema)
    incoming_table=order_fields_in_table(incoming_table, merged_schema)

    # # Combine the tables
    combined_table = pa.concat_tables([current_table, incoming_table])
    
    return combined_table


def create_empty_table(schema: pa.Schema, columns: list = None) -> pa.Table:
    """
    Creates an empty PyArrow table with the same schema as the dataset or specific columns.

    Args:
        schema (pa.Schema): The schema of the dataset to mimic in the empty generator.
        columns (list, optional): List of column names to include in the empty table. Defaults to None.

    Returns:
        pa.Table: An empty PyArrow table with the specified schema.
    """
    # If specific columns are provided, filter the schema to include only those columns
    if columns:
        schema = pa.schema([field for field in schema if field.name in columns])

    # Create an empty table with the derived schema
    empty_table = pa.Table.from_pydict({field.name: [] for field in schema}, schema=schema)

    return empty_table



def create_empty_batch_generator(schema: pa.Schema, columns: list = None):
    """
    Returns an empty generator that yields nothing.

    Args:
        schema (pa.Schema): The schema of the dataset to mimic in the empty generator.
        columns (list, optional): List of column names to include in the empty table. Defaults to None.
    Yields:
        pa.RecordBatch: Empty record batches with the specified schema.
    """
    if columns:
        schema = pa.schema([field for field in schema if field.name in columns])
    yield pa.RecordBatch.from_pydict({field.name: [] for field in schema}, schema=schema)





from typing import List, Callable

import pyarrow as pa
import copy

dummy_struct_type=pa.struct([('dummy_field', pa.int16())])

def replace_empty_structs(struct_type, dummy_struct_type=pa.struct([('dummy_field', pa.int16())])):
    # If the struct is empty, return the dummy struct
    if len(struct_type)==0:
        return dummy_struct_type

    field_list=[]
    # Iterate over the fields in the struct
    for field in struct_type:
        field_name=field.name
        # Handles the determination of the field type
        if pa.types.is_struct(field.type):
            # Handles empty structs.
            if len(field.flatten())==0:
                field_type=dummy_struct_type
            # Handles nested structs.
            else:
                field_type=replace_empty_structs(field.type)
        else:
            field_type=field.type

        field_list.append((field_name,field_type))
        print(field_list)

    return pa.struct(field_list)




def align_struct_fields(original_array: pa.Array, new_type: pa.DataType) -> pa.Array:
    """
    Aligns the fields of an array to match the new type, filling missing fields with null values.
    Handles nested struct types recursively and replaces empty structs with a dummy struct.

    Args:
        original_array (pa.Array): The original array (can be a struct or any other type).
        new_type (pa.DataType): The target type to align to.

    Returns:
        pa.Array: The aligned array.
    """
    # Combine chunks if necessary
    if isinstance(original_array, pa.ChunkedArray):
        original_array = pa.concat_arrays(original_array.chunks)
    
    # If the new type is not a struct, just cast and return
    if not isinstance(new_type, pa.StructType):
        return original_array.cast(new_type)
    
    original_type = original_array.type
    original_fields_dict = {field.name: i for i, field in enumerate(original_type)}
    
    # Build a new struct with all the required fields
    new_arrays = []
    for field in new_type:
        if field.name in original_fields_dict:
            # If the field exists, align it recursively
            original_field = original_array.field(original_fields_dict[field.name])
            aligned_field = align_struct_fields(original_field, field.type)
            
            # Check if the aligned field is an empty struct
            if isinstance(field.type, pa.StructType) and len(field.type) == 0:
                # Replace empty struct with dummy struct
                dummy_array = pa.array([{'dummy_field': None}] * len(original_array), type=dummy_struct_type)
                new_arrays.append(dummy_array)
            else:
                new_arrays.append(aligned_field)
        else:
            # If the field doesn't exist, fill with nulls
            if isinstance(field.type, pa.StructType) and len(field.type) == 0:
                # For empty struct fields, use the dummy struct
                dummy_array = pa.array([{'dummy_field': None}] * len(original_array), type=dummy_struct_type)
                new_arrays.append(dummy_array)
            else:
                null_array = pa.nulls(len(original_array), field.type)
                new_arrays.append(null_array)
    
    return pa.StructArray.from_arrays(new_arrays, fields=new_type)
# Example usage
def create_example_data():
    # Original data with nested structs, including an empty struct
    data = [
        {'a': 1, 'b': {'x': 10, 'y': 20}, 'c': {}},
        {'a': 2, 'b': {'x': 30, 'y': 40}, 'c': {}},
    ]
    original_array = pa.array(data)

    # New schema with additional fields and an empty struct
    new_schema = pa.struct([
        ('a', pa.int64()),
        ('b', pa.struct([
            ('x', pa.int64()),
            ('y', pa.int64()),
            ('z', pa.int64()),  # New field
        ])),
        ('c', pa.struct([])),  # Empty struct
        ('d', pa.int64()),  # New field
    ])
    new_schema=replace_empty_structs(new_schema)
    return original_array, new_schema

def replace_empty_struct_in_schema(schema, dummy_struct_type=pa.struct([('dummy_field', pa.int16())])):
    callbacks_kwargs=[{'dummy_struct_type':dummy_struct_type}]
    new_schema=iterate_over_schema_fields(schema, callabacks = [replace_empty_structs_in_field], 
                                        callbacks_kwargs=callbacks_kwargs)
    return new_schema

def replace_empty_structs_in_field(field, dummy_struct_type=pa.struct([('dummy_field', pa.int16())])):
    field_name=field.name
    if pa.types.is_struct(field.type):
        new_type=replace_empty_structs(field.type,dummy_struct_type)
    else:
        new_type=field.type
    return pa.field(field_name, new_type)



def iterate_over_schema_fields(schema, callabacks: List[Callable], callbacks_kwargs: List[dict]=None):
    if len(callabacks)==0:
        return schema

    new_fields=[]
    for field in schema:
        new_field=copy.deepcopy(field)
        for i,callback in enumerate(callabacks):
            if callbacks_kwargs:
                callback_kwargs=callbacks_kwargs[i]
            else:
                callback_kwargs={}
            new_field=callback(new_field, **callback_kwargs)

        new_fields.append(new_field)
    return pa.schema(new_fields)

def test_callback(field):
    field_name=field.name
    if pa.types.is_struct(field.type):
        new_type=replace_empty_structs(field.type)
        return pa.field(field_name, new_type)
    else:
        return field


def main():
    # original_array, new_schema = create_example_data()
    # aligned_array = align_struct_fields(original_array, new_schema)
    
    # print("Original array:")
    # print(original_array)
    # print("\nAligned array:")
    # df=aligned_array.to_pandas()

    # print(df.iloc[0]['c'])
    # print(aligned_array.to_pandas())

    # data = [
    #     { 'c': {}},
    #     { 'c': {}},
    # ]
    data = [
        {'a': 1, 'b': {'x': 10, 'y': 20}, 'c': {}},
        {'a': 2, 'b': {'x': 30, 'y': 40}, 'c': {}},
    ]
    # print(data)
    table=pa.Table.from_pylist(data)
    print(table.schema)
    new_schema=replace_empty_struct_in_schema(schema=table.schema)
    print(new_schema)

    # result=iterate_over_schema_fields(table.schema, callabacks = [replace_empty_structs_in_field])
    # print(result)

if __name__ == "__main__":
    main()


import logging
import pyarrow as pa
import pyarrow.dataset as ds




logger=logging.getLogger('parquetdb')
logger.setLevel(logging.DEBUG)
ch = logging.StreamHandler()
ch.setLevel(logging.DEBUG)
formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
ch.setFormatter(formatter)
logger.addHandler(ch)

def create_empty_table(schema: pa.Schema, columns: list = None) -> pa.Table:
    """
    Creates an empty PyArrow table with the same schema as the dataset or specific columns.

    Args:
        schema (pa.Schema): The schema of the dataset to mimic in the empty generator.
        columns (list, optional): List of column names to include in the empty table. Defaults to None.

    Returns:
        pa.Table: An empty PyArrow table with the specified schema.
    """
    # If specific columns are provided, filter the schema to include only those columns
    if columns:
        schema = pa.schema([field for field in schema if field.name in columns])

    # Create an empty table with the derived schema
    empty_table = pa.Table.from_pydict({field.name: [] for field in schema}, schema=schema)

    return empty_table



def create_empty_batch_generator(schema: pa.Schema, columns: list = None):
    """
    Returns an empty generator that yields nothing.

    Args:
        schema (pa.Schema): The schema of the dataset to mimic in the empty generator.
        columns (list, optional): List of column names to include in the empty table. Defaults to None.
    Yields:
        pa.RecordBatch: Empty record batches with the specified schema.
    """
    if columns:
        schema = pa.schema([field for field in schema if field.name in columns])
    yield pa.RecordBatch.from_pydict({field.name: [] for field in schema}, schema=schema)



if __name__ == "__main__":
    
    from parquetdb import ParquetDatasetDB,ParquetDB
    data = [
        {'a': 1, 'b': {'x': 10, 'y': 20}, 'c': {}},
        {'a': 2, 'b': {'x': 30, 'y': 40}, 'c': {}},
    ]
    # print(data)
    

    db.create(data, dataset_name='main')
    # dataset = ds.dataset(db.dataset_dir, format="parquet")
    # schema = dataset.schema
    # Create an empty table# Example usage
    
    # db.create(data)
    
    table=db.read(columns=['a','b'])
    print(table.num_rows)
    table=db.read(columns=['t'])
    print(table.num_rows)
    
    generator=db.read(columns=['a','b'], batch_size=10)
    for table in generator:
        print(table)
        print(table.num_rows)
    generator=db.read(columns=['t'], batch_size=10)
    for table in generator:
        print(table)
        print(table.num_rows)
    
    # print(table.num_rows)
    # df=table.to_pandas()
    # print(df.columns)
    # print(df.head())    
    
    # dataset = ds.dataset(db.dataset_dir, format="parquet")
    # try:
    #     table=dataset.to_table(columns=['t'],filter=None)
    #     df=table.to_pandas()
    # except pa.lib.ArrowInvalid:
    #     print("Exception loading table")
    # print(df.columns)
    # print(df.head())
    # print(df.shape)
    
    # empty_table = create_empty_table(schema=schema, columns=['a','b'])
    # print(empty_table.schema)
    # print(empty_table.num_rows)
    
    # generator=create_empty_batch_generator(schema=schema, columns=['t'])
    # # print(dir(generator))
    # for table in generator:
    #     print(table)
    #     print(table.num_rows)
    # # print(empty_table)
    
    print('reeeee')



import logging
import random
import os
import time
import tempfile
import matplotlib as mpl
mpl.use('Agg')
import matplotlib.pyplot as plt
from glob import glob
import json

import pyarrow as pa

from parquetdb import ParquetDB

logger=logging.getLogger('parquetdb')
logger.setLevel(logging.DEBUG)
ch = logging.StreamHandler()
ch.setLevel(logging.DEBUG)
formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
ch.setFormatter(formatter)
logger.addHandler(ch)


materials_dir = "C:/Users/lllang/Desktop/Current_Projects/ParquetDB/data/raw/materials_data"
files = glob(os.path.join(materials_dir, '*.json'))

# Load materials data from JSON files
materials_data = []
for file in files:
    with open(file, 'r') as f:
        materials_data.append(json.load(f))


save_dir = 'C:/Users/lllang/Desktop/Current_Projects/ParquetDB/data/raw/ParquetDB'
db = ParquetDB(db_path=save_dir)
db.drop_table('main_test')

# data_list=[]
# for i in range(1000):
#     data = random.choices(materials_data,k=1)[0]
#     data_list.append(data)
# db.create(data=data_list, table_name='main_test', max_rows_per_file=100, min_rows_per_group=0, max_rows_per_group=100)

data_list=[]
for i in range(100):
    data = materials_data[0]
    data_list.append(data)

db.create(data=data_list, table_name='main_test', max_rows_per_file=100, min_rows_per_group=0, max_rows_per_group=100)

data_list=[]
for i in range(100):
    data = materials_data[1]
    data_list.append(data)

db.create(data=data_list, table_name='main_test', max_rows_per_file=100, min_rows_per_group=0, max_rows_per_group=100)


data_list=[]
for i in range(100):
    data = materials_data[2]
    data_list.append(data)

db.create(data=data_list, table_name='main_test', max_rows_per_file=100, min_rows_per_group=0, max_rows_per_group=100)



table=db.read(table_name='main_test')

# schema=db.get_schema('main_test')
# # print(schema)

# print(detect_struct_types(schema))


df=table.to_pandas()
print(df.columns)
print(df.head())
print(df.tail())
print(df.shape)
print(df['composition'].iloc[-1])
print(df['structure'].iloc[-1])


db.update([{'id':299, 'nsites': 100, 'composition': {'Ca':1.0}, 'structure': {'lattice': {'matrix': [[1, 0, 0], [0, 1, 0], [0, 0, 1]]}}}], table_name='main_test')
table=db.read(table_name='main_test')
df=table.to_pandas()
print(df.columns)
print(df.head())
print(df.tail())
print(df.shape)

print(df['structure'].iloc[-1])
print(df['composition'].iloc[-1])
print(df['nsites'].iloc[-1])
# print(df['field1'])
# print(df['field2'])
# print(df.shape)


################################################################################################
# Testing Update functionality
################################################################################################

# data_list=[]
# for i in range(1000):
#     data={}
#     data['id']=i
#     data['field5']=5
#     data_list.append(data)

# data_list=[{'id':1, 'field5':5, 'field6':6},
#            {'id':2, 'field5':5, 'field6':6},
#            {'id':1000, 'field5':5, 'field8':6},
#            ]

# db.update(data_list)

# table=db.read(ids=[1,2,1000],  table_name='main', columns=['id','field5','field6', 'field8'])
# print(table)


################################################################################################
# Testing Delete functionality
################################################################################################

# Deleting rows in table
# db.delete(ids=[1,3,1000,2])
# db.delete(ids=[2])
# table=db.read(ids=[1,3,1000,2],  table_name='main', columns=['id','field5','field6', 'field8'])
# print(table)


# Trying to delete a row that does not exist
# db.delete(ids=[10000000])

################################################################################################
# Testing Read functionality
################################################################################################
# # Basic read
# table=db.read( table_name='main')
# df=table.to_pandas()
# print(df.head())
# print(df.tail())
# print(df.shape)

# # Testing ids
# df=db.read(ids=[1,2], table_name='main')
# print(df.head())
# print(df.tail())
# print(df.shape)

# # # Testing columns include_cols
# df=db.read(table_name='main', columns=['volume'], include_cols=True)
# print(df.head())
# print(df.tail())
# print(df.shape)

# # # Testing columns include_cols False
# df=db.read(table_name='main', columns=['volume'], include_cols=False)
# print(df.head())
# print(df.tail())
# print(df.shape)

# # # # Testing columns filter_func
# df=db.read(table_name='main', filter_func=test_filter_func)
# print(df.head())
# print(df.tail())
# print(df.shape)

# # # # Testing columns filter_args
# df=db.read(table_name='main', filter_args=('band_gap',  '==', 1.593))
# print(df.head())
# print(df.tail())
# print(df.shape)

########################################################################################


from typing import List, Callable, Dict

import pyarrow as pa
import copy


class SchemaModifier:
    def __init__(self):
        self.callbacks = []
        self.callbacks_kwargs = []

    def add_callback(self, callback: Callable, callback_kwargs: Dict = None):
        """Add a callback function along with its keyword arguments."""
        self.callbacks.append(callback)
        if callback_kwargs is None:
            callback_kwargs = {}
        self.callbacks_kwargs.append(callback_kwargs)

    def modify_schema(self, schema: pa.Schema):
        """Iterate over schema fields and apply the callbacks."""
        return self._iterate_over_schema_fields(schema, self.callbacks, self.callbacks_kwargs)

    def _iterate_over_schema_fields(self, schema: pa.Schema, callbacks: List[Callable], callbacks_kwargs: List[Dict] = None):
        """Internal function to iterate over schema fields and apply each callback."""
        if len(callbacks) == 0:
            return schema

        new_fields = []
        for field in schema:
            new_field = copy.deepcopy(field)
            for i, callback in enumerate(callbacks):
                if callbacks_kwargs:
                    callback_kwargs = callbacks_kwargs[i]
                else:
                    callback_kwargs = {}
                new_field = callback(new_field, **callback_kwargs)

            new_fields.append(new_field)
        return pa.schema(new_fields)

def replace_empty_structs(struct_type, dummy_struct_type=pa.struct([('dummy_field', pa.int16())])):
    # If the struct is empty, return the dummy struct
    if len(struct_type)==0:
        return dummy_struct_type

    field_list=[]
    # Iterate over the fields in the struct
    for field in struct_type:
        field_name=field.name
        # Handles the determination of the field type
        if pa.types.is_struct(field.type):
            # Handles empty structs.
            if len(field.flatten())==0:
                field_type=dummy_struct_type
            # Handles nested structs.
            else:
                field_type=replace_empty_structs(field.type)
        else:
            field_type=field.type

        field_list.append((field_name,field_type))

    return pa.struct(field_list)


def replace_empty_structs_in_field(field, dummy_struct_type=pa.struct([('dummy_field', pa.int16())])):
    if pa.types.is_struct(field.type):
        field_name=field.name
        if pa.types.is_struct(field.type):
            new_type=replace_empty_structs(field.type,dummy_struct_type)
        else:
            new_type=field.type

    else:
        field_name=field.name
        new_type=field.type
    
    return pa.field(field_name, new_type)


data = [
        {'a': 1, 'b': {'x': 10, 'y': 20}, 'c': {}},
        {'a': 2, 'b': {'x': 30, 'y': 40}, 'c': {}},
    ]
# print(data)
table=pa.Table.from_pylist(data)

if __name__ == "__main__":
    # Initialize the SchemaModifier
    modifier = SchemaModifier()

    # Add the callback with arguments
    modifier.add_callback(replace_empty_structs_in_field, {'dummy_struct_type': pa.struct([('dummy_field', pa.int16())])})

    # Define a schema and modify it
    original_schema = pa.schema([pa.field('example_field', pa.struct([]))])
    new_schema = modifier.modify_schema(original_schema)

    print(new_schema)


from tempfile import mkdtemp
import logging

from parquetdb import ParquetDB
import pyarrow.compute as pc

logger=logging.getLogger('parquetdb')
logger.setLevel(logging.DEBUG)
ch = logging.StreamHandler()
ch.setLevel(logging.DEBUG)
formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
ch.setFormatter(formatter)
logger.addHandler(ch)


initial = {
		"global": [
			{
				"key": b"\xb4_lise_schema_version",
				"value": b"\x00",
			},
			{"key": b"\xabmain_branch", "value": b"\xa5trunk"},
			{"key": b"\xa6branch", "value": b"\xa5trunk"},
			{"key": b"\xa4turn", "value": b"\x00"},
			{"key": b"\xa4tick", "value": b"\x00"},
			{"key": b"\xa8language", "value": b"\xa3eng"},
		]
	}

dir = mkdtemp()
db = ParquetDB(dir)

def set_global(key: bytes, value: bytes):
	try:
		id_ = db.read(dataset_name="global", filters=[pc.field("key") == key], columns=["id"])["id"][0].as_py()
		db.update(
			{"id": id_, "key": key, "value": value}, "global"
		)
	except:
		db.create({"key": key, "value": value}, "global")


set_global(b"\xa6branch", b"\xa5tronc")
set_global(b"\xa4turn", b"\x00")

df = db.read(dataset_name="global").to_pandas()
print(df)
set_global(b"\xa6branch", b"\x00")
print('-'*200)

df = db.read(dataset_name="global").to_pandas()
print(df)


from tempfile import mkdtemp

from parquetdb import ParquetDatasetDB
import pyarrow.compute as pc

initial = {
		"global": [
			{
				"key": b"\xb4_lise_schema_version",
				"value": b"\x00",
			},
			{"key": b"\xabmain_branch", "value": b"\xa5trunk"},
			{"key": b"\xa6branch", "value": b"\xa5trunk"},
			{"key": b"\xa4turn", "value": b"\x00"},
			{"key": b"\xa4tick", "value": b"\x00"},
			{"key": b"\xa8language", "value": b"\xa3eng"},
		]
	}

dir = mkdtemp()
dir = "C:/Users/lllang/Desktop/Current_Projects/ParquetDB/data/raw/ParquetDB_Dev"
db = ParquetDatasetDB(dataset_name="global", dir=dir, n_cores=1)

def set_global(key: bytes, value: bytes):
	try:
		id_ = db.read(filters=[pc.field("key") == key], columns=["id"])["id"][0].as_py()
		db.update(
			{"id": id_, "key": key, "value": value}, "global"
		)
	except:
		db.create({"key": key, "value": value})


set_global(b"\xa6branch", b"\xa5tronc")
set_global(b"\xa4turn", b"\x00")

set_global(b"\xa8language", b"\xa3eng")


set_global(b"\xa4turn",b"\xa5tronc")

# db.create(initial, "global")

df = db.read().to_pandas()
print(df)


import logging
import random
import os
import shutil
import time
import tempfile
import matplotlib as mpl
mpl.use('Agg')
import matplotlib.pyplot as plt
from glob import glob
import json

import pyarrow as pa

from parquetdb import ParquetDB, ParquetDatasetDB
from parquetdb.pyarrow_utils import align_table, merge_schemas, merge_structs, combine_tables

logger=logging.getLogger('parquetdb')
logger.setLevel(logging.DEBUG)
ch = logging.StreamHandler()
ch.setLevel(logging.DEBUG)
formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
ch.setFormatter(formatter)
logger.addHandler(ch)

##########################################################################################################################
# materials_dir = "C:/Users/lllang/Desktop/Current_Projects/ParquetDB/data/raw/materials_data"
# files = glob(os.path.join(materials_dir, '*.json'))

# # Load materials data from JSON files
# materials_data = []
# for file in files:
#     with open(file, 'r') as f:
#         materials_data.append(json.load(f))
##########################################################################################################################

import pyarrow as pa


# def process_type(data_type):
#     if pa.types.is_struct(data_type):
#         fields = [field for field in struct_type]

#         if len(fields) == 0:
#             # Create a dummy field for empty structs
#             dummy_field = pa.field('dummy', pa.null())
#             return pa.struct([dummy_field])
#         else:
#             # Recursively process each field in the struct
#             new_fields = [pa.field(field.name, process_type(field.type), field.nullable, field.metadata)
#                           for field in fields]
#             return pa.struct(new_fields)
#     elif pa.types.is_list(data_type):
#         # Process the value type of the list
#         value_type = process_type(data_type.value_type)
#         return pa.list_(value_type)
#     elif pa.types.is_large_list(data_type):
#         value_type = process_type(data_type.value_type)
#         return pa.large_list(value_type)
#     elif pa.types.is_map(data_type):
#         # Process the key and item types of the map
#         key_type = process_type(data_type.key_type)
#         item_type = process_type(data_type.item_type)
#         return pa.map_(key_type, item_type)
#     elif pa.types.is_union(data_type):
#         # Process each child type of the union
#         new_fields = [pa.field(field.name, process_type(field.type), field.nullable, field.metadata)
#                       for field in data_type]
#         return pa.union(new_fields, data_type.mode)
#     else:
#         # Return the data type as is for other types
#         return data_type

# def process_schema(schema):
#     # Recursively process the schema to modify data types
#     new_fields = [pa.field(field.name, process_type(field.type), field.nullable, field.metadata)
#                   for field in schema]
#     return pa.schema(new_fields, metadata=schema.metadata)

# def process_array(array, old_type, new_type):
#     if old_type == new_type:
#         return array
#     elif pa.types.is_struct(old_type):
#         if pa.types.is_struct(new_type):
#             old_fields = {field.name: field for field in old_type}
#             new_fields = {field.name: field for field in new_type}
#             child_arrays = []
#             for field_name in new_fields:
#                 if field_name in old_fields:
#                     old_child_array = array.field(field_name)
#                     child_array = process_array(old_child_array,
#                                                 old_fields[field_name].type,
#                                                 new_fields[field_name].type)
#                 else:
#                     # Create a null array for the new dummy field
#                     length = len(array)
#                     child_array = pa.nulls(length, type=new_fields[field_name].type)
#                 child_arrays.append(child_array)
#             # Create a new StructArray with updated fields
#             return pa.StructArray.from_arrays(child_arrays, fields=list(new_fields.values()))
#         else:
#             raise TypeError("Type mismatch: old type is struct, new type is not struct")
#     elif pa.types.is_list(old_type):
#         if pa.types.is_list(new_type):
#             old_value_type = old_type.value_type
#             new_value_type = new_type.value_type
#             value_array = array.values
#             new_value_array = process_array(value_array, old_value_type, new_value_type)
#             return pa.ListArray.from_arrays(array.offsets, new_value_array)
#         else:
#             raise TypeError("Type mismatch: old type is list, new type is not list")
#     elif pa.types.is_large_list(old_type):
#         if pa.types.is_large_list(new_type):
#             old_value_type = old_type.value_type
#             new_value_type = new_type.value_type
#             value_array = array.values
#             new_value_array = process_array(value_array, old_value_type, new_value_type)
#             return pa.LargeListArray.from_arrays(array.offsets, new_value_array)
#         else:
#             raise TypeError("Type mismatch: old type is large list, new type is not large list")
#     else:
#         # For other types, attempt to cast
#         try:
#             return array.cast(new_type)
#         except:
#             raise TypeError(f"Cannot cast array of type {old_type} to {new_type}")

# def process_chunked_array(chunked_array, old_type, new_type):
#     new_chunks = [process_array(chunk, old_type, new_type) for chunk in chunked_array.chunks]
#     return pa.chunked_array(new_chunks, type=new_type)

# def add_dummy_field_to_empty_structs(table):
#     # Process the schema to get the new schema
#     new_schema = process_schema(table.schema)
#     new_columns = []
#     for i, field in enumerate(table.schema):
#         old_type = field.type
#         new_type = new_schema.field(i).type
#         chunked_array = table.column(i)
#         # Process each column's data to match the new schema
#         new_chunked_array = process_chunked_array(chunked_array, old_type, new_type)
#         new_columns.append(new_chunked_array)
#     # Create a new table with the updated schema and data
#     new_table = pa.Table.from_arrays(new_columns, schema=new_schema)
#     return new_table

save_dir = 'C:/Users/lllang/Desktop/Current_Projects/ParquetDB/data/raw/ParquetDB_Dev'

temp_dir = tempfile.mkdtemp()
db = ParquetDatasetDB(dataset_name='dev', dir=save_dir, n_cores=1)
# db.drop_table('dev')



data=[{'field1':{
            'field-11': {'field-21': {}, 'field-22': '5'} , 
            'field-12': {}}, 
        'field2':1}
    ]
struct_type = pa.struct({'x': pa.int32(), 'y': pa.string()})

field_names= [field.name for field in struct_type]
# print(struct_type.names)
print(field_names)
table=pa.Table.from_pylist(data)


schema=table.schema
# def find_struct_fields(schema, find_empty_structs=False):
#     struct_fields=[]
#     empty_fields=[]
#     for field in schema:
#         if pa.types.is_struct(field.type):
#             print(field.name, field.type)
#             print(field)
            
#             flatten_field = field.flatten()
            
#             if len(flatten_field)==0:
#                 empty_fields.append(field)
#             struct_fields.append(field)
            
#             nested_struct_fields,nested_empty_fields=find_struct_fields(flatten_field)
#             if nested_struct_fields:
#                 struct_fields.extend(nested_struct_fields)
#             if nested_empty_fields:
#                 empty_fields.extend(nested_empty_fields)

#     return struct_fields, empty_fields

def find_struct_fields(schema):
    struct_fields=[]
    for field in schema:
        if pa.types.is_struct(field.type):
            struct_fields.append(field)
            flatten_field = field.flatten()
            nested_struct_fields=find_struct_fields(flatten_field)
            if nested_struct_fields:
                struct_fields.extend(nested_struct_fields)

    return struct_fields

def find_empty_structs(schema):
    empty_fields=[]
    struct_fields=find_struct_fields(schema)
    for field in struct_fields:
        if len(field.flatten())==0:
            empty_fields.append(field)
    return empty_fields

def add_dummy_field_to_schemas_with_empty_structs(table):
    schema=table.schema
    print(schema.field('field1.field-11'))
    empty_fields=find_empty_structs(schema)
    print(empty_fields)
    for field in empty_fields:
        field_name=field.name
        nested_field_names=field_name.split('.')
        nested_field=None
        for i,nest_filed_name in enumerate(nested_field_names):
            # print(i)
            # print(dir(field.type))
            # if i==len(nested_field_names)-1:
            #     continue
            if nested_field is None:
                nested_field=schema.field(nest_filed_name)
            else:
                nested_field=nested_field.type.field(nest_filed_name)
            # if nested_field is None:
            #     nested_field=schema.field(nest_filed_name)
            # else:
            #     nested_field=nested_field.field(nest_filed_name)
        # nested_field = 
        print(nested_field)
        print(nested_field_names)
    # return struct_fields

# for field in schema:
#     print(field.name, field.type)

# empty_fields=find_empty_structs(schema)

# empty_field=empty_fields[0]
# print(empty_field)
# nested_field_names=empty_field.name.split('.')
# field=schema.field(nested_field_names[0])
# print(field)
dummy_struct_type=pa.struct([('dummy_field', pa.int16())])
def replace_empty_structs(struct_type):
    field_list=[]
    for field in struct_type:
        field_name=field.name
        # Handles the determination of the field type
        if pa.types.is_struct(field.type):
            # Handles empty structs.
            if len(field.flatten())==0:
                field_type=dummy_struct_type
            # Handles nested structs.
            else:
                field_type=replace_empty_structs(field.type)
        else:
            field_type=field.type

        field_list.append((field_name,field_type))
    return pa.struct(field_list)

def replace_empty_structs_in_schema(schema):
    field_list=[]
    for field in schema:
        field_name=field.name
        field_type=field.type
        if pa.types.is_struct(field.type):
            field_type=replace_empty_structs(field.type)
        else:
            field_type=field.type
        field_list.append((field_name,field_type))
    return pa.schema(field_list)


new_schema=replace_empty_structs_in_schema(schema)


def add_null_values_for_missing_nested_fields(chunked_array, new_struct_type):

    new_struct_names= [field.name for field in struct_type]
    current_names= [field.name for field in chunked_array.type]
    # table_schema = table.schema
    # current_names= [field.name for field in table_schema.field(field_name)]

    # field_type=field.type

    field_names_missing = list(set(new_struct_names) - set(current_names))
    
    # Add missing fields with null values
    for new_field_name in field_names_missing:
        field_type = new_struct_type.field(new_field_name).type
        if pa.types.is_struct(field_type):
            add_null_values_for_missing_nested_fields(chunked_array, new_struct_type)
        # null_array = pa.nulls(table.num_rows, type=field_type)
        # table = table.append_column(new_field_name, null_array)
    
    # return field
# column_table=table.select(columns=['field1'])
# column=table.column('field1')
# print(column)
# print(type(column))
# print(column.type)



####################################################################################
data=[{'field1':{
            'field-11': {'field-21': {'field-31':1}, 'field-22': '5'} , 
            'field-12': {'field-21':1}}, 
        'field2':1},
        {'field1':{
            'field-11': {'field-21': {'field-32':1}, 'field-22': '5'} , 
            'field-12': {'field-23':1}}, 
        'field2':1},
        {'field1':{
            'field-11': {'field-21': {'field-32':1}, 'field-22': '5'} , 
            'field-12': {'field-23':1}}, 
        'field2':1}
    ]

table_1=pa.Table.from_pylist(data)


data=[{'field1':{
            'field-11': {'field-21': {'field-32':1}, 'field-22': '5'} , 
            'field-12': {'field-23':1}}, 
        'field2':1}]

table_2=pa.Table.from_pylist(data)
new_schema=merge_schemas(current_schema=table_1.schema, incoming_schema=table_2.schema)
# new_schema=table_1.schema
new_table_2=align_table(table, new_schema)
print(new_table_2.to_pandas())


# # print(select_column)
# # print(type(select_column))
# # # print(table_1.select(columns=['field1']))
# # def align_struct_chunked_arrays(chunk_array_1, chunk_array_2):
# #     field_type_1=chunk_array_1.type
# #     field_type_2=chunk_array_2.type

# #     field_names_1=[field.name for field in field_type_1]
# #     field_names_2=[field.name for field in field_type_2]

# #     flattened_array_1=chunk_array_1.flatten()
# #     flattened_array_2=chunk_array_2.flatten()
    
# #     missing_field_name=list(set(field_names_1) - set(field_names_2))
# #     for missing_field in missing_field_name:
#     # for field in struct_type:
#     # print(combined_chunked_array)
#     # for field in struct_type:
#     #     field_name=field.name
#     #     # Handles the determination of the field type
#     #     if pa.types.is_struct(field.type):
#     #         # Handles empty structs.
#     #         if len(field.flatten())==0:
#     #             field_type=dummy_struct_type
#     #         # Handles nested structs.
#     #         else:
#     #             field_type=replace_empty_structs(field.type)
#     #     else:
#     #         field_type=field.type



# new_schema=table_1.schema
# new_type=merge_schemas(current_schema=table_1.column('field1').type, 
#                          incoming_schema=table_2.column('field1').type)
# print(new_type)


# def align_struct_fields(original_array: pa.ChunkedArray, new_type: pa.StructType) -> pa.ChunkedArray:

#     # Combine chunks if necessary
#     original_array = pa.concat_arrays(original_array.chunks) if isinstance(original_array, pa.ChunkedArray) else original_array
    
#     original_type = original_array.type
#     original_fields_dict = {field.name: i for i, field in enumerate(original_type)}
    
#     # Build a new struct with all the required fields
#     new_arrays = []
#     for field in new_type:
#         if field.name in original_fields_dict:
#             # If the field exists, keep the original field
#             new_arrays.append(original_array.field(original_fields_dict[field.name]))
#         else:
#             # Otherwise, fill with nulls
#             null_array = pa.nulls(len(original_array), field.type)
#             new_arrays.append(null_array)
    
#     return pa.StructArray.from_arrays(new_arrays, fields=new_type)

# # align_struct_fields(original_array=table_1.column('field1'), new_type=new_type)

# print(table_2.column('field1'))

# # new_schema
# # align_struct_columns(column_table=table_1.select('field1'), 
# #                     chunk_array_2=table_2.column('field1'))

# # new_
# chunk_array_1=table_1.column('field1')
# print(type(chunk_array_1))

# print(type(chunk_array_1[0]))
# print(chunk_array_1[0])

# # print(len(table_1.column('field1')))
# # print(type(table_1.column('field1')))
# # align_struct_chunked_arrays(chunk_array_1=table_1.column('field1'), 
# #                             chunk_array_2=table_2.column('field1'))

# Create pyarrow tables
# table_1 = pa.Table.from_pylist(data_1)
# table_2 = pa.Table.from_pylist(data_2)

# Extract schemas
# schema_1 = table_1.schema
# schema_2 = table_2.schema

# def align_struct_fields(original_array: pa.Array, new_type: pa.DataType) -> pa.Array:
#     """
#     Aligns the fields of an array to match the new schema, filling missing fields with null values.
#     Handles nested struct types recursively.

#     Args:
#         original_array (pa.Array): The original array (can be a struct or any other type).
#         new_type (pa.DataType): The target type to align to.

#     Returns:
#         pa.Array: The aligned array.
#     """
#     # Combine chunks if necessary
#     if isinstance(original_array, pa.ChunkedArray):
#         original_array = pa.concat_arrays(original_array.chunks)
    
#     # If the new type is not a struct, just cast and return
#     if not isinstance(new_type, pa.StructType):
#         return original_array.cast(new_type)
    
#     original_type = original_array.type
#     original_fields_dict = {field.name: i for i, field in enumerate(original_type)}
    
#     # Build a new struct with all the required fields
#     new_arrays = []
#     for field in new_type:
#         if field.name in original_fields_dict:
#             # If the field exists, align it recursively
#             original_field = original_array.field(original_fields_dict[field.name])
#             aligned_field = align_struct_fields(original_field, field.type)
#             new_arrays.append(aligned_field)
#         else:
#             # Otherwise, fill with nulls
#             null_array = pa.nulls(len(original_array), field.type)
#             new_arrays.append(null_array)
    
#     return pa.StructArray.from_arrays(new_arrays, fields=new_type)


# aligned_array = align_struct_fields(table_1.column('field1'), new_schema.field('field1').type)

# print("Original array:")
# print(table_1.column('field1'))
# print("\nAligned array:")
# print(aligned_array.to_pandas())


from parquetdb.core.parquet_datasetdb import ParquetDatasetDB

db = ParquetDatasetDB(dataset_name='dev', dir='C:/Users/lllang/Desktop/Current_Projects/ParquetDB/data/raw/ParquetDB_Dev', n_cores=1)

data = [
    {'name': 'Judy', 'age': 29}
        ]
db.create(data)

# Add new data with an additional field
new_data = [
    {'name': 'Karl', 'occupation': 'Engineer'}
]
db.create(new_data)

db.create(data)


result = db.read()
df = result.to_pandas()

print(df.columns)
print(df.head())


import logging
import os
import shutil
from functools import partial
from glob import glob
from multiprocessing import Pool
import traceback
from typing import List, Union

import pandas as pd
import pyarrow as pa
import pyarrow.compute as pc
import pyarrow.dataset as ds
import pyarrow.parquet as pq

from parquetdb.utils.general_utils import timeit, is_directory_empty
from parquetdb.utils import pyarrow_utils

# Logger setup
logger = logging.getLogger(__name__)


def get_field_names(filepath, columns=None, include_cols=True):
    """
    Gets the field names from a parquet file.

    Args:
        filepath (str): The path to the parquet file.
        columns (list): A list of column names to include.
        include_cols (bool): If True, includes all columns except the ones listed in columns.

    Returns:
        list: A list of column names.
    """
    if not include_cols:
        metadata = pq.read_metadata(filepath)
        all_columns = []
        for filed_schema in metadata.schema:
            
            # Only want top column names
            max_defintion_level=filed_schema.max_definition_level
            if max_defintion_level!=1:
                continue

            all_columns.append(filed_schema.name)

        columns = [col for col in all_columns if col not in columns]
    return columns

class ParquetDatasetDB:
    def __init__(self, dataset_name, dir='', n_cores=8):
        """
        Initializes the ParquetDBDataset object.

        Args:
            db_path (str): The path to the root directory of the database.
            n_cores (int): The number of CPU cores to be used for parallel processing.
        """
        self.dir = dir
        self.dataset_name=dataset_name
        self.dataset_dir=os.path.join(self.dir,self.dataset_name)
        self.tmp_dir=os.path.join(self.dir,'tmp')
        self.n_cores = n_cores
        self.basename_template = f'{dataset_name}_{{i}}.parquet'


        os.makedirs(self.tmp_dir, exist_ok=True)
        os.makedirs(self.dataset_dir, exist_ok=True)

        
        self.output_formats=['batch_generator','table','dataset']
        self.reserved_dataset_names=['tmp']

        logger.info(f"dir: {self.dir}")
        logger.info(f"reserved_dataset_names: {self.reserved_dataset_names}")
        logger.info(f"n_cores: {self.n_cores}")
        logger.info(f"output_formats: {self.output_formats}")

    
    @timeit
    def create(self, data:Union[List[dict],dict,pd.DataFrame],  
               batch_size:int=None,
               max_rows_per_file=10000,
               min_rows_per_group=0,
               max_rows_per_group=10000,
               schema=None,
               metadata=None,
               **kwargs):
        """
        Adds new data to the database.

        Args:
            data (dict or list of dicts): The data to be added to the database. 
                This must contain
            dataset_name (str): The name of the table to add the data to.
            batch_size (int): The batch size. 
                If provided, create will return a generator that yields batches of data.
            max_rows_per_file (int): The maximum number of rows per file.
            min_rows_per_group (int): The minimum number of rows per group.
            max_rows_per_group (int): The maximum number of rows per group.
            schema (pyarrow.Schema): The schema of the incoming table.
            metadata (dict): Metadata to be added to the table.
            **kwargs: Additional keyword arguments to pass to the create function.
        """
        
        os.makedirs(self.dataset_dir, exist_ok=True)
        
        # Prepare the data and field data
        data_list=self._validate_data(data)
        new_ids = self._get_new_ids( data_list)

        # Create the incoming table
        incoming_table=pa.Table.from_pylist(data_list, schema=schema, metadata=metadata)
        incoming_table=incoming_table.append_column(pa.field('id', pa.int64()), [new_ids])

        # If this is the first table, save it directly
        if is_directory_empty(self.dataset_dir):
            incoming_table=pyarrow_utils.replace_empty_structs_in_table(incoming_table)
            incoming_save_path = self._get_save_path()
            pq.write_table(incoming_table, incoming_save_path)
            return None

        # Write tmp files for the original table
        # This is done to ensure that the original table is not deleted 
        # if there is an exception during the writing process
        self._write_tmp_files()

        # Align the incoming and original schemas
        try:
            # Merge schemas
            incoming_schema = incoming_table.schema
            current_schema = self.get_schema()
            merged_schema=pyarrow_utils.merge_schemas(current_schema, incoming_schema)

            # Align file schemas with merged schema
            current_files = glob(os.path.join(self.dataset_dir, f'{self.dataset_name}_*.parquet'))
            for current_file in current_files:
                current_table = pq.read_table(current_file)

                # Align current schema to match incoming
                updated_table=pyarrow_utils.align_table(current_table, merged_schema)

                pq.write_table(updated_table, current_file)

            # Align incoming tbale to match merged schema
            incoming_table=pyarrow_utils.align_table(incoming_table, merged_schema)

        except Exception as e:
            logger.exception(f"exception aligning schemas: {e}")
            logger.info("Restoring original files")
            self._restore_tmp_files()
        
        # Save the incoming table
        incoming_save_path = self._get_save_path()
        pq.write_table(incoming_table, incoming_save_path)

        self._finalize_dataset(incoming_table.schema, batch_size, max_rows_per_file, 
                                min_rows_per_group, max_rows_per_group, **kwargs)
        return None

    @timeit
    def read(
        self,
        ids: List[int] = None,
        columns: List[str] = None,
        include_cols: bool = True,
        filters: List[pc.Expression] = None,
        output_format: str = 'table',
        batch_size: int = None,
        load_kwargs: dict = None
        ) -> Union[pa.Table, pa.dataset.Scanner]:
        """
        Reads data from the database.

        Args:
            ids (list): A list of IDs to read. If None, reads all data.
            dataset_name (str): The name of the table to read data from.
            columns (list): A list of columns to include in the returned data. By default, all columns are included.
            include_cols (bool): If True, includes the only the fields listed in columns
                If False, includes all fields except the ones listed in columns.
            filters (List): A list of fliters to apply to the data.
            It should operate on a dataframe and return the modifies dataframe
            batch_size (int): The batch size. 
                If provided, read will return a generator that yields batches of data.
            load_kwargs (dict): Additional keyword arguments to pass to Dataset.to_table or Dataset.to_batches.

        Returns:
            pandas.DataFrame or list: The data read from the database. If deserialize_data is True,
            returns a list of dictionaries with their 'id's. Otherwise, returns the DataFrame with serialized data.
        """

        # Check if the table name is in the list of table names
        table_path=os.path.join(self.dataset_dir, f'{self.dataset_name}.parquet')

        if columns:
            columns=get_field_names(table_path, columns=columns, include_cols=include_cols)

        if filters is None:
            filters = []
            
        if batch_size:
            logger.info(f"Found batch_size: {batch_size}. Setting output_format to batch_generator")
            output_format='batch_generator'
            
        # Build filter expression
        filter_expression = self._build_filter_expression(ids, filters)

        data = self._load_data(columns=columns, filter=filter_expression, 
                               batch_size=batch_size, output_format=output_format, load_kwargs=load_kwargs)
        return data
    
    @timeit
    def update(self, data: Union[List[dict], dict, pd.DataFrame], field_type_dict=None):
        """
        Updates data in the database.

        Args:
            data (dict or list of dicts or pandas.DataFrame): The data to be updated.
                Each dict should have an 'id' key corresponding to the record to update.
            dataset_name (str): The name of the table to update data in.
            field_type_dict (dict): A dictionary where the keys are the field names and the values are the new field types.

            **kwargs: Additional keyword arguments.

        Raises:
            ValueError: If new fields are found in the update data that do not exist in the schema.
        """

        # Data processing and validation.
        data_list = self._validate_data(data)

        # Process update data and get id_data_dict_map and update schema
        id_data_dict_map, update_schema = self._process_update_data(data_list, field_type_dict)
        
        logger.info(f"Dataset directory: {self.dataset_dir}")
        current_files=self._get_current_files()

        # Iterate over the original files
        self._write_tmp_files()

        for i_file, current_file in enumerate(current_files):
            current_table = pq.read_table(current_file)

            try:
                updated_table = self._update_table(current_table, id_data_dict_map, update_schema)
            except Exception as e:
                logger.exception(f"exception updating {current_file}")
                # logger.exception(f"exception updating {current_file}: {e}")
                # If something goes wrong, restore the original file
                self._restore_tmp_files()
                break

            pq.write_table(updated_table, current_file)
            logger.info(f"Updated {current_file} with {updated_table.shape}")
        logger.info(f"Updated {self.dataset_name} table.")

    @timeit
    def delete(self, ids:List[int]):
        """
        Deletes data from the database.

        Args:
            ids (list): A list of IDs to delete.
            dataset_name (str): The name of the table to delete data from.

        Returns:
            None
        """
        ids=set(ids)

        logger.info(f"Deleting data from {self.dataset_name}")

        main_id_column=self._load_data(columns=['id'], output_format='table')

        # Check if any of the IDs to delete exist in the table
        filtered_table = main_id_column.filter( pc.field('id').isin(ids) )
        if filtered_table.num_rows==0:
            logger.info(f"No data found to delete.")
            return None
        
        # Dataset directory
        logger.info(f"Dataset directory: {self.dataset_dir}")
        current_files = self._get_current_files()

        # Backup current files in case of failure
        self._write_tmp_files()

        for current_file in current_files:
            filename=os.path.basename(current_file)

            current_table=pq.read_table(current_file)

            # Delete the IDs from the table
            try:
                updated_table = self._delete_ids_from_table(ids, current_table)
            except Exception as e:
                logger.exception(f"exception processing {current_file}: {e}")
                # If something goes wrong, restore the original file
                self._restore_tmp_files()
                raise e
            
            # Saving the new table with the deleted IDs
            pq.write_table(updated_table, current_file)

            n_rows_deleted=current_table.shape[0]-updated_table.shape[0]
            logger.info(f"Deleted {n_rows_deleted} Indices from {filename}. The shape is now {updated_table.shape}")

        logger.info(f"Updated {self.dataset_name} table.")

    @timeit
    def update_schema(self, field_dict:dict=None, schema:pa.Schema=None):
        """
        Updates the schema of the table.

        Args:
            dataset_name (str): The name of the table to update the schema of.
            field_dict (dict): A dictionary where the keys are the field names and the values are the new field types.
            schema (pyarrow.Schema): The new schema for the table.

        """
        # Dataset directory
        current_files=self._get_current_files()

        # Backup current files in case of failure
        self._write_tmp_files()

        for current_file in current_files:
            filename=os.path.basename(current_file)
            current_table=pq.read_table(current_file)

            updated_table=self._update_dataset_schema(current_table, schema, field_dict)

            try:
                pq.write_table(updated_table, current_file)
            except Exception as e:
                logger.exception(f"exception processing {current_file}: {e}")
                # If something goes wrong, restore the original file
                self._restore_tmp_files()
                break

            logger.info(f"Updated {filename} with {current_table.shape}")

        logger.info(f"Updated Fields in {self.dataset_name} table.")

    def get_schema(self):
        """Get the schema of a table.

        Args:
            dataset_name (str): The name of the table.

        Returns:
            pyarrow.Schema: The schema of the table.
        """
        schema = self._load_data(output_format='dataset').schema
        return schema
    
    def get_metadata(self):
        """Get the metadata of a table.
        
        Args:
            dataset_name (str): The name of the table.

        Returns:
            dict: The metadata of the table.
        """
        if not self.dataset_exists():
            raise ValueError(f"Dataset {self.dataset_name} does not exist.")
        schema = self.get_schema()
        return schema.metadata
    
    def set_metadata(self, metadata:dict):
        """Set the metadata of a table.

        Args:
            dataset_name (str): The name of the table.
            metadata (dict): The metadata to set.
        """
        # Update metadata in schema and rewrite Parquet files
        self.update_schema(schema=pa.schema(self.get_schema().fields, metadata=metadata))

    def dataset_exists(self, dataset_name:str=None):
        """Checks if a table exists.

        Returns:
            bool: True if the table exists, False otherwise.
        """
        
        if dataset_name:
            dataset_dir=os.path.join(self.dir,dataset_name)
            return os.path.exists(dataset_dir)
        else:
            return len(self._get_current_files()) != 0

    def drop_dataset(self):
        """
        Drops a table. by removing the table directory.

        Args:
            dataset_name (str): The name of the table to drop.
        """
      
        if os.path.exists(self.dataset_dir):
            shutil.rmtree(self.dataset_dir)
            logger.info(f"Table {self.dataset_name} has been dropped.")
        else:
            logger.warning(f"Table {self.dataset_name} does not exist.")
    
    def rename_dataset(self, new_name:str):
        """
        Renames a table.

        Args:
            old_name (str): The current name of the table.
            new_name (str): The new name of the table.
        """
        if not self.dataset_exists():
            raise ValueError(f"Dataset {self.dataset_name} does not exist.")
        if new_name in self.reserved_dataset_names:
            raise ValueError(f"Cannot rename to reserved table name: {new_name}")
        old_dir = os.path.join(self.dataset_dir)
        old_name=self.dataset_name
        

        # Rename all files in the old directory
        old_filepaths = glob(os.path.join(old_dir, f'{old_name}_*.parquet'))
        for old_filepath in old_filepaths:
            filename=os.path.basename(old_filepath)
            file_index=filename.split('.')[0].replace('_','')
            new_filepath = os.path.join(old_dir, f'{new_name}_{file_index}.parquet')
            os.rename(old_filepath, new_filepath)

        # Finally, rename the directory
        new_dir = os.path.join(self.dir, new_name)
        os.rename(old_dir, new_dir)

        self.dataset_name=new_name
        self.dataset_dir=new_dir

        logger.info(f"Table {old_name} has been renamed to {new_name}.")

    def copy_dataset(self, dest_name: str, overwrite: bool = False):
        """
        Copies a table to a new table.

        Args:
            source_table (str): The name of the source table.
            dest_table (str): The name of the destination table.
        """
        if overwrite and self.dataset_exists(dest_name):
            shutil.rmtree(os.path.join(self.dir, dest_name))
        elif self.dataset_exists(dest_name):
            raise ValueError(f"Dataset {dest_name} already exists.")
        if dest_name in self.reserved_dataset_names:
            raise ValueError(f"Cannot copy to reserved table name: {dest_name}")
        
        source_dir = self.dataset_dir
        source_name=self.dataset_name
        dest_dir = os.path.join(self.dir, dest_name)
        
        os.makedirs(dest_dir, exist_ok=True)

        # Rename all files in the old directory
        old_filepaths = glob(os.path.join(source_dir, f'{source_name}_*.parquet'))
        for old_filepath in old_filepaths:
            filename=os.path.basename(old_filepath)
            file_index=filename.split('.')[0].replace('_','')
            new_filepath = os.path.join(dest_dir, f'{dest_name}_{file_index}.parquet')
            shutil.copyfile(old_filepath, new_filepath)

        logger.info(f"Table {source_name} has been copied to {dest_name}.")

    def optimize_dataset(self, 
                    max_rows_per_file=10000,
                    min_rows_per_group=0,
                    max_rows_per_group=10000,
                    batch_size=None,
                    **kwargs):
        """
        Optimizes the table by merging small Parquet files.

        Args:
            dataset_name (str): The name of the table to optimize.
            max_rows_per_file (int): The maximum number of rows per file.
            min_rows_per_group (int): The minimum number of rows per group.
            max_rows_per_group (int): The maximum number of rows per group.
            batch_size (int): The batch size.
            **kwargs: Additional keyword arguments to pass to the pq.write_to_dataset function.
        """

        # Read the entire dataset either in batches or as a whole
        schema=None
        if batch_size:
            table = self._load_data(output_format='batch_generator', batch_size=batch_size)
            schema=self.get_schema()
        else:
            table = self._load_data(output_format='table')
            
        
        # Backup current files in case of failure
        self._write_tmp_files()

        try:
            # Write optimized dataset
            pq.write_dataset(
                table,
                self.dataset_dir,
                basename_template=self.basename_template,
                schema=schema,
                format="parquet",
                max_rows_per_file=max_rows_per_file,
                min_rows_per_group=0,
                max_rows_per_group=10000,
                existing_data_behavior='overwrite_or_ignore'
                **kwargs
            )
            logger.info(f"Optimized table {self.dataset_name}.")
        except Exception as e:
            logger.exception(f"exception optimizing table {self.dataset_name}: {e}")
            logger.info("Restoring original files")
            self._restore_tmp_files()

    def export_dataset(self, file_path: str, format: str = 'csv'):
        """
        Exports the table to a specified file format.

        Args:
            dataset_name (str): The name of the table to export.
            file_path (str): The file path to export the data to.
            format (str): The format to export ('csv', 'json').
        """
        table = self._load_data(output_format='table')
        if format == 'csv':
            df = table.to_pandas()
            df.to_csv(file_path, index=False)
        elif format == 'json':
            df = table.to_pandas()
            df.to_json(file_path, orient='records', lines=True)
        else:
            raise ValueError(f"Unsupported export format: {format}")
        logger.info(f"Exported table {self.dataset_name} to {file_path} as {format}.")

    def export_partitioned_dataset(self,
                                   export_dir: str, 
                                   partitioning,
                                   partitioning_flavor=None,
                                   batch_size: int = None, 
                                   **kwargs):
        """
        This method exports a partitioned dataset to a specified file format.

        Args:
            dataset_name (str): The name of the table to export.
            export_dir (str): The directory to export the data to.
            partitioning (dict): The partitioning to use for the dataset.
            partitioning_flavor (str): The partitioning flavor to use.
            batch_size (int): The batch size.
            **kwargs: Additional keyword arguments to pass to the pq.write_to_dataset function.

        """

        # Read the entire dataset either in batches or as a whole
        schema=None
        if batch_size:
            table = self._load_data( output_format='batch_generator', batch_size=batch_size)
            schema=self.get_schema()
        else:
            table = self._load_data(output_format='table')

        pq.write_to_dataset(
                table,
                export_dir,
                schema=schema,
                partitioning=partitioning,
                partitioning_flavor=partitioning_flavor,
                format="parquet",
                **kwargs
            )

    def import_dataset(self, file_path: str, format: str = 'csv', **kwargs):
        """
        Imports a table from a specified file format.

        Args:
            file_path (str): The file path to import the data from.
            dataset_name (str): The name of the table to import the data into.
            format (str): The format to import ('csv', 'json').
            **kwargs: Additional keyword arguments to pass to the create function.
        """
        if format == 'csv':
            df = pd.read_csv(file_path, **kwargs)
        elif format == 'json':
            df = pd.read_json(file_path, **kwargs)
        else:
            raise ValueError(f"Unsupported import format: {format}")
        self.create(data=df.to_dict(orient='records'))
        logger.info(f"Imported data from {file_path} into table {self.dataset_name}.")

    def merge_datasets(self, source_tables: List[str], dest_table: str):
        raise NotImplementedError
    
    def backup_database(self, backup_path: str):
        """
        Creates a backup of the database.

        Args:
            backup_path (str): The path where the backup will be stored.
        """
        shutil.copytree(self.dataset_dir, backup_path)
        logger.info(f"Database backed up to {backup_path}.")

    def restore_database(self, backup_path: str):
        """
        Restores the database from a backup.

        Args:
            backup_path (str): The path to the backup to restore from.
        """
        if os.path.exists(self.dataset_dir):
            shutil.rmtree(self.dataset_dir)
        shutil.copytree(backup_path, self.dataset_dir)
        logger.info(f"Database restored from {backup_path}.")

    def _load_data(self, columns:List[str]=None, 
                   filter:List[pc.Expression]=None, 
                   batch_size:int=None, 
                   output_format:str='table',
                   load_tmp:bool=False,
                   load_kwargs:dict=None):
        """
        This method loads the data in the database. It can either load the data as a PyArrow Table, PyArrow Dataset, PyArrow generator.
        
        Parameters
        ----------
        
        columns : List[str], optional
            A list of column names to load. If None, loads all columns. Defaults to None.
        filter : List[pc.Expression], optional
            A list of filters to apply to the data. Defaults to None.
        batch_size : int, optional
            The batch size to use for loading data in batches. Defaults to None.
        output_format : str, optional
            The output format to use for loading data. Defaults to 'table'.
        load_tmp : bool, optional
            Whether to load data from the temporary directory. Defaults to False.
        load_kwargs : dict, optional
            Additional keyword arguments to pass to Dataset.to_table or Dataset.to_batches. Defaults to None.

        Returns
        -------
        Union[pa.Table, pa.dataset.Scanner]
            The loaded data as a PyArrow Table, PyArrow Dataset, or PyArrow generator.
        
        """
        if load_kwargs is None:
            load_kwargs={}
            
        dataset_dir=self.dataset_dir
        if load_tmp:
            dataset_dir=self.tmp_dir
            
        logger.info(f"Loading data from {dataset_dir}")
        logger.info(f"Loading only columns: {columns}")
        logger.info(f"Using filter: {filter}")

        dataset = ds.dataset(dataset_dir, format="parquet")
        if output_format=='batch_generator':
            return self._load_batches(dataset, batch_size, columns, filter, **load_kwargs)
        elif output_format=='table':
            return self._load_table(dataset, columns, filter, **load_kwargs)
        elif output_format=='dataset':
            return dataset
        else:
            raise ValueError(f"output_format must be one of the following: {self.output_formats}")
    
    def _load_batches(self, dataset, batch_size, columns:List[str]=None, filter:List[pc.Expression]=None, **kwargs):
        if batch_size is None:
            raise ValueError("batch_size must be provided when output_format is batch_generator")
        logger.info(f"Loading in batches")
        
        try:
            generator=dataset.to_batches(columns=columns,filter=filter, batch_size=batch_size, **kwargs)
        except Exception as e:
            logger.debug(f"Error loading table: {e}. Returning empty table")
            generator=pyarrow_utils.create_empty_batch_generator(schema=dataset.schema, columns=columns)
        return generator
    
    def _load_table(self, dataset, columns:List[str]=None, filter:List[pc.Expression]=None, **kwargs):
        logger.info(f"Loading data in one table")
        try:
            table=dataset.to_table(columns=columns,filter=filter,**kwargs)
        except Exception as e:
            logger.debug(f"Error loading table: {e}. Returning empty table")
            table=pyarrow_utils.create_empty_table(schema=dataset.schema, columns=columns)
        return table
    
    def _get_new_ids(self, data_list:List[dict]):
  
        if is_directory_empty(self.dataset_dir):
            start_id = 0
        else:
            table = self._load_data(columns=['id'],output_format='table')
            max_val=pc.max(table.column('id')).as_py()
            start_id = max_val + 1  # Start from the next available ID
    
        # Create a list of new IDs
        new_ids = list(range(start_id, start_id + len(data_list)))
        return new_ids
    
    def _build_filter_expression(self, ids: List[int], filters: List[pc.Expression]):
        """Helper function to build the filter expression."""
        final_filters = []
        
        # Add ID filter if provided
        if ids:
            id_filter = pc.field('id').isin(ids)
            final_filters.append(id_filter)

        # Append custom filters
        final_filters.extend(filters)

        # Combine filters into a single filter expression
        if not final_filters:
            return None
        
        filter_expression = final_filters[0]
        for filter_expr in final_filters[1:]:
            filter_expression = filter_expression & filter_expr

        return filter_expression

    def _get_save_path(self):
        """Determine the path to save the incoming table."""

        n_files = len(glob(os.path.join(self.dataset_dir, f'{self.dataset_name}_*.parquet')))
        
        if n_files == 0:
            return os.path.join(self.dataset_dir, f'{self.dataset_name}_0.parquet')
        return os.path.join(self.dataset_dir, f'{self.dataset_name}_{n_files}.parquet')
    
    def add_null_columns_for_missing_fields(self, table, new_schema):
        """
        This function adds null columns for any fields that are missing in the new schema.

        Args:
            table (pa.Table): The table to add null columns to.
            new_schema (pa.Schema): The schema of the new table.

        Returns:
            pa.Table: The table with null columns added for missing fields.
        """
        table_schema = table.schema
        field_names_missing = list(set(new_schema.names) - set(table_schema.names))
        for new_field_name in field_names_missing:
            field_type = new_schema.field(new_field_name).type
            null_array = pa.nulls(table.shape[0], type=field_type)
            table = table.append_column(new_field_name, null_array)
        return table
    
    def _finalize_dataset(self,
                        schema:pa.Schema=None,
                        batch_size:int=None, 
                        max_rows_per_file:int=10000, 
                        min_rows_per_group:int=0, 
                        max_rows_per_group:int=10000, 
                        **kwargs):
        """Finalize the creation by writing the data to disk."""

        output_format='table'
        if batch_size:
            output_format='batch_generator'
        else:
            schema=None

        self._write_tmp_files()

        try:
            final_table = self._load_data(batch_size=batch_size, output_format=output_format, load_tmp=True)
        except pa.lib.ArrowNotImplementedError as e:
            raise ValueError("The incoming data does not match the schema of the existing data.") from e

        try:
            logger.info(f"Writing final table to {self.dataset_dir}")
            shutil.rmtree(self.dataset_dir)
            
            ds.write_dataset(final_table, self.dataset_dir, basename_template=self.basename_template, schema=schema,
                            format="parquet", max_partitions=kwargs.get('max_partitions', 1024),
                            max_open_files=kwargs.get('max_open_files', 1024), max_rows_per_file=max_rows_per_file, 
                            min_rows_per_group=min_rows_per_group, max_rows_per_group=max_rows_per_group,
                            existing_data_behavior='overwrite_or_ignore')
            
        except Exception as e:
            logger.exception(f"exception writing final table to {self.dataset_dir}: {e}")
            logger.info("Restoring original files")
            self._restore_tmp_files()

    def _process_update_data(self, data_list: List[dict], field_type_dict=None):
        """Process and validate incoming data, return update dictionary, incoming field names, and inferred types."""
        if field_type_dict is None:
            field_type_dict={}
        
        id_data_dict_map = {}
        incoming_field_names = set()
        infered_types=None
        
        main_id_column=self._load_data(columns=['id'], 
                                       output_format='table')['id'].to_pylist()
        # Create id_data_dict_map, get incoming field names, and their infered types
        for data_dict in data_list:
            data_id = data_dict.get('id', None)

            # Check if the id is in the main table
            self._validate_id(data_id, main_id_column)

            # Add data_dict to id_data_dict_map
            id_data_dict_map[data_id] = data_dict

            # Get incoming field names
            incoming_field_names.update(data_dict.keys())

            # Determine inferred types for each field (skip 'id')
            infered_types=self._infer_pyarrow_types(data_dict)

        # Get the current schema
        current_schema = self.get_schema()
        current_names=set(current_schema.names)

        # Resolve data infered types with current schema
        infered_types=self._resolve_infered_field_types(current_schema, infered_types)

        merged_names=list(incoming_field_names.union(current_names) - set(['id']))
        update_schema = pa.schema([(field_name, infered_types[field_name]) for field_name in merged_names])

        return id_data_dict_map, update_schema

    def _infer_pyarrow_types(self, data_dict: dict):
        infered_types = {}
        for key, value in data_dict.items():
            if key != 'id':
                infered_types[key] = pa.infer_type([value])
        return infered_types

    def _validate_id(self, id: List[dict], id_column: List[int]):
        if id is None:
            raise ValueError("Each data dict must have an 'id' key.")
        if id not in id_column:
            raise ValueError(f"The id {id} is not in the main table. It may have been deleted or is incorrect.")
        return None
    
    def _resolve_infered_field_types(self, schema: pa.Schema, infered_types: dict):
        logger.info(f"Checking infered field types vs original field types")
        for field_name in schema.names:
            type=schema.field(field_name).type
            infered_types[field_name]=type

        logger.info(f"Validating field types: {infered_types}")
        return infered_types
    
    def _update_table(self, current_table: str, 
                     id_data_dict_map: dict, 
                     update_schema: pa.Schema):
        """Update a single Parquet file with new data."""

        # Add new columns to the current table
        current_table=pyarrow_utils.replace_empty_structs_in_table(current_table)
        update_schema = pyarrow_utils.merge_schemas(current_table.schema, update_schema)

        logger.debug(f"Aligning table with update schema")
        updated_table = pyarrow_utils.align_table(current_table, update_schema)
        logger.debug(f"Aligned table with update schema")
        # Update existing records in the current table
        current_ids_list = current_table['id'].to_pylist()
        id_to_index = {id: index for index, id in enumerate(current_ids_list)}
        # Iterate over the columns in the table
        for column_idx, column in enumerate(updated_table.itercolumns()):
            column_name = column._name
            if column_name == 'id':  # Skip 'id' column
                continue
            
            # Get current column array
            column_array = column.to_pylist()

            # Iterate over the update entries for the current column
            for id, data_dict in id_data_dict_map.items():
                
                # If the id is in the the current table 
                # and the field_nmae is in the update dict, update the column
                if id in id_to_index and column_name in data_dict:
                    
                    update_value = data_dict[column_name]

                    # If the update value is None for a column, skip it. 
                    # This can happen when a field is added to another 
                    # file but not the current one

                    if update_value is not None:  # Only update non-None values
                        # if isinstance(update_value, dict):
                        #     column_array[id_to_index[id]].update(update_value)
                        # else:
                        #     column_array[id_to_index[id]] = update_value

                        index = id_to_index[id]
                        current_value = column_array[index]

                        # Perform a deep update if it's a nested structure
                        column_array[index] = deep_update(current_value, update_value)

            # Update the column in the table
            field = updated_table.schema.field(column_name)

            # Ensure the column array is of the correct type
            column_array=pa.array(column_array, type=field.type)
            updated_table = updated_table.set_column(column_idx, field, column_array)

        return updated_table

    def _delete_ids_from_table(self, ids, table):
        """Delete IDs from a table."""
        # Get the original ids
        current_ids_list=set(table['id'].to_pylist())

        # Get the ids that are in the current table
        ids_in_table = current_ids_list & ids

        # Applying the negative id filter
        new_table = table.filter(~pc.field('id').isin(ids_in_table))
        return new_table
    
    def _update_dataset_schema(self, current_table, schema=None, field_dict=None):
        """Update the schema of a table."""
        # Check if the table name is in the list of table names

        current_schema=current_table.schema
        current_field_names=current_table.column_names

        if field_dict:
            updated_schema=current_schema
            for field_name, new_field in field_dict.items():
                field_index=current_schema.get_field_index(field_name)

                if field_name in current_field_names:
                    updated_schema=updated_schema.set(field_index, new_field)

        if schema:
            updated_schema=schema

        pylist=current_table.to_pylist()
        new_table=pa.Table.from_pylist(pylist, schema=updated_schema)

        return new_table

    def _write_tmp_files(self, tmp_dir=None):
        if tmp_dir is None:
            tmp_dir=self.tmp_dir
        shutil.rmtree(self.tmp_dir)
        os.makedirs(self.tmp_dir, exist_ok=True)

        current_filepaths=glob(os.path.join(self.dataset_dir,f'{self.dataset_name}_*.parquet'))
        for i_file, current_filepath in enumerate(current_filepaths):
            basename=os.path.basename(current_filepath)
            
            tmp_filepath = os.path.join(self.tmp_dir, basename)
            shutil.copyfile(current_filepath, tmp_filepath)

    def _restore_tmp_files(self, tmp_dir=None):
        if tmp_dir is None:
            tmp_dir=self.tmp_dir
 
        tmp_filepaths=glob(os.path.join(self.tmp_dir,f'{self.dataset_name}_*.parquet'))
        for i_file, tmp_filepath in enumerate(tmp_filepaths):
            basename=os.path.basename(tmp_filepath)
            current_filepath = os.path.join(self.dataset_dir, basename)
            shutil.copyfile(tmp_filepath, current_filepath)
            os.remove(tmp_filepath)
    
    def _validate_data(self, data):
        if isinstance(data, dict):
            data_list = [data]
        elif isinstance(data, list):
            data_list = data
        elif isinstance(data, pd.DataFrame):
            data_list = data.to_dict(orient='records')
        elif data is None:
            data_list = None
        else:
            raise Typeexception("Data must be a dictionary or a list of dictionaries.")
        return data_list
    
    def _get_current_files(self):

        return glob(os.path.join(self.dataset_dir, f'{self.dataset_name}_*.parquet'))

def add_dummy_field_to_empty_structs(table: pa.Table, dummy_field_name="dummy_field", dummy_field_type=pa.int32()) -> pa.Table:
    """
    Search for struct types with no child fields and add a dummy field to those structs.

    Parameters:
    - table: A PyArrow Table to check for empty struct types.
    - dummy_field_name: Name of the dummy field to add.
    - dummy_field_type: DataType of the dummy field (default is pa.int32).

    Returns:
    - A new PyArrow Table with the updated schema.
    """
    schema = table.schema
    new_fields = []

    # Traverse the schema fields
    for field in schema:
        if pa.types.is_struct(field.type):
            struct_type = field.type
            if len(struct_type) == 0:
                # Empty struct found, add a dummy field to it
                new_struct_type = pa.struct([pa.field(dummy_field_name, dummy_field_type)])
                new_fields.append(pa.field(field.name, new_struct_type))
            else:
                # Keep the existing struct field with its children
                new_fields.append(field)
        else:
            # Keep non-struct fields as is
            new_fields.append(field)

    # Create new schema with updated fields
    new_schema = pa.schema(new_fields)

    # Reconstruct the table with the updated schema
    # Assuming you can reconstruct the data based on new schema
    return table.cast(new_schema)

def deep_update(original_value, update_value):
    """
    Recursively updates original_value with the corresponding values in update_value.
    If update_value contains nested dictionaries or lists, it updates them recursively.
    
    Args:
        original_value: The original value to be updated (could be a dict, list, or primitive).
        update_value: The update value (could be a dict, list, or primitive).
    
    Returns:
        Updated original_value with changes from update_value.
    """
    # If both are dictionaries, update recursively
    if isinstance(original_value, dict) and isinstance(update_value, dict):
        for key, value in update_value.items():
            if key in original_value:
                original_value[key] = deep_update(original_value[key], value)
            else:
                original_value[key] = value
        return original_value
    
    # # If both are lists, update recursively by index
    # elif isinstance(original_value, list) and isinstance(update_value, list):
    #     # Extend the original list if update_value has more items
    #     for i in range(len(update_value)):
    #         if i < len(original_value):
    #             original_value[i] = deep_update(original_value[i], update_value[i])
    #         else:
    #             original_value.append(update_value[i])
    #     return original_value
    
    # Otherwise, just return the update value (for primitive types or if the types differ)
    return update_value
